#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.09.22 19:00:00                  #
# ================================================== #

import re

from PySide6.QtCore import Qt, QObject, Signal, Slot, QEvent, QCoreApplication, QEventLoop, QUrl
from PySide6.QtWebChannel import QWebChannel
from PySide6.QtWebEngineCore import QWebEngineSettings, QWebEnginePage
from PySide6.QtWebEngineWidgets import QWebEngineView
from PySide6.QtGui import QAction, QIcon
from PySide6.QtWidgets import QMenu

from pygpt_net.item.ctx import CtxMeta
from pygpt_net.core.text.web_finder import WebFinder
from pygpt_net.utils import trans


class HtmlOutput(QWebEngineView):
    def __init__(self, window=None):
        """
        HTML output (Webkit)

        :param window: main window
        """
        super(HtmlOutput, self).__init__(window)
        self.window = window
        self.finder = WebFinder(window, self)
        self.loadFinished.connect(self.on_page_loaded)
        self.customContextMenuRequested.connect(self.on_context_menu)
        self.signals = WebEngineSignals()
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.plain = ""
        self.html_content = ""
        self.meta = None
        self.tab = None
        self.setMouseTracking(True)

        # OpenGL widgets
        self._glwidget = None
        self._glwidget_filter_installed = False
        self._unloaded = False
        self._destroyed = False

        self.installEventFilter(self)

    def _detach_gl_event_filter(self):
        """Detach OpenGL widget event filter if installed"""
        if self._glwidget and self._glwidget_filter_installed:
            try:
                self._glwidget.removeEventFilter(self)
            except Exception as e:
                self._on_delete_failed(e)
        self._glwidget = None
        self._glwidget_filter_installed = False

    def _on_delete_failed(self, e):
        """
        Handle delete failure

        :param e: Exception instance
        """
        pass
        self.window.core.debug.log(e)

    def unload(self):
        """Unload the current page and free resources"""
        try:
            self.hide()
            p = self.page()
            p.triggerAction(QWebEnginePage.Stop)
            p.setUrl(QUrl("about:blank"))
            p.history().clear()
            p.setLifecycleState(QWebEnginePage.LifecycleState.Discarded)
        except Exception as e:
            self._on_delete_failed(e)
        finally:
            self._unloaded = True

    def on_delete(self):
        """Clean up on delete"""
        if self._destroyed:
            return
        if not self._unloaded:
            self.unload()

        self.hide()
        self._detach_gl_event_filter()

        if self.finder:
            try:
                self.finder.disconnect()
            except Exception as e:
                self._on_delete_failed(e)
            finally:
                self.finder = None

        self.tab = None

        # disconnect signals
        try:
            self.loadFinished.disconnect(self.on_page_loaded)
            self.customContextMenuRequested.disconnect(self.on_context_menu)
            self.signals.save_as.disconnect()
            self.signals.audio_read.disconnect()
        except Exception as e:
            self._on_delete_failed(e)

        page = self.page()

        try:
            page.triggerAction(QWebEnginePage.Stop)
        except Exception as e:
            self._on_delete_failed(e)

        try:
            page.setUrl(QUrl("about:blank"))
        except Exception as e:
            self._on_delete_failed(e)

        try:
            page.history().clear()
        except Exception as e:
            self._on_delete_failed(e)

        try:
            page.setLifecycleState(QWebEnginePage.LifecycleState.Discarded)
        except Exception as e:
            self._on_delete_failed(e)

        try:
            if hasattr(page, "setWebChannel"):
                page.setWebChannel(None)
        except Exception as e:
            self._on_delete_failed(e)

        prof = None
        try:
            prof = page.profile()
        except Exception as e:
            self._on_delete_failed(e)

        """
        if prof is not None:
            try:
                prof.deleteLater()
            except Exception as e:
                self._on_delete_failed(e)
        """

        try:
            self.deleteLater()
        except Exception as e:
            self._on_delete_failed(e)

        try:
            QCoreApplication.sendPostedEvents(None, QEvent.DeferredDelete)
            QCoreApplication.processEvents(QEventLoop.AllEvents, 50)
        except Exception as e:
            self._on_delete_failed(e)

        self._destroyed = True

    def set_tab(self, tab):
        """
        Set tab

        :param tab: Tab
        """
        self.tab = tab

    def set_meta(self, meta: CtxMeta):
        """
        Assign ctx meta

        :param meta: context meta
        """
        self.meta = meta

    def set_plaintext(self, text: str):
        """
        Set plain text

        :param text: text
        """
        self.plain = text

    def set_html_content(self, html: str):
        """
        Set HTML content

        :param html: HTML content
        """
        self.html_content = html

        # set page HTML
        self.setHtml(html, baseUrl="file://")

    def get_html_content(self) -> str:
        """
        Get HTML content

        :return: HTML content
        """
        return self.html_content

    def on_context_menu(self, position):
        """
        Context menu event

        :param position: position
        """
        menu = QMenu(self)
        selected_text = ""
        is_selection = self.page().hasSelection()
        if is_selection:
            selected_text = self.get_selected_text()

        if is_selection:
            # copy
            action = QAction(QIcon(":/icons/copy.svg"), trans('action.copy'), self)
            action.triggered.connect(self.copy_selected_text)
            menu.addAction(action)

            # audio read
            action = QAction(QIcon(":/icons/volume.svg"), trans('text.context_menu.audio.read'), self)
            action.triggered.connect(
                lambda: self.signals.audio_read.emit(selected_text)
            )
            menu.addAction(action)

            # copy to
            copy_to_menu = self.window.ui.context_menu.get_copy_to_menu(menu, selected_text)
            menu.addMenu(copy_to_menu)

            # save as (selected)
            action = QAction(QIcon(":/icons/save.svg"), trans('action.save_selection_as'), self)
            action.triggered.connect(
                lambda: self.signals.save_as.emit(selected_text, 'txt')
            )
            menu.addAction(action)
        else:
            # select all
            action = QAction(QIcon(":/icons/copy.svg"), trans('action.select_all'), self)
            action.triggered.connect(self.select_all_text)
            menu.addAction(action)

            # save as (all) - plain
            action = QAction(QIcon(":/icons/save.svg"), trans('action.save_as') + " (text)", self)
            action.triggered.connect(
                lambda: self.signals.save_as.emit(re.sub(r'\n{2,}', '\n\n', self.plain), 'txt')
            )
            menu.addAction(action)

            # save as (all) - html
            action = QAction(QIcon(":/icons/save.svg"), trans('action.save_as') + " (html)", self)
            action.triggered.connect(
                lambda: self.signals.save_as.emit(re.sub(r'\n{2,}', '\n\n', self.html_content), 'html')
            )
            menu.addAction(action)

        action = QAction(QIcon(":/icons/search.svg"), trans('text.context_menu.find'), self)
        action.triggered.connect(self.find_open)
        # action.setShortcut(QKeySequence("Ctrl+F"))
        menu.addAction(action)

        menu.exec_(self.mapToGlobal(position))

    def update_zoom(self):
        """Update zoom from config"""
        if self.window.core.config.has("zoom"):
            self.page().setZoomFactor(self.window.core.config.get("zoom"))

    def get_zoom_value(self) -> float:
        """
        Get zoom value

        :return: zoom value
        """
        return self.page().zoomFactor()

    def reset_current_content(self):
        """Reset current content"""
        self.plain = ""
        self.html_content = ""

    def update_current_content(self):
        """Update current content"""
        self.page().runJavaScript("document.getElementById('container').outerHTML", 0, self.set_plaintext)
        self.page().runJavaScript("document.documentElement.innerHTML", 0, self.set_html_content)

    def on_page_loaded(self, success):
        """
        On page loaded

        :param success: True if loaded successfully
        """
        pass

    def get_selected_text(self) -> str:
        """
        Get selected text

        :return: selected text
        """
        return self.page().selectedText()

    def copy_selected_text(self):
        """Copy selected text"""
        self.page().triggerAction(QWebEnginePage.Copy)

    def select_all_text(self):
        """Select all text"""
        self.page().triggerAction(QWebEnginePage.SelectAll)

    def unselect_text(self):
        """Unselect text"""
        self.page().triggerAction(QWebEnginePage.Unselect)

    def find_open(self):
        """Open find dialog"""
        self.window.controller.finder.open(self.finder)

    def on_update(self):
        """On content update"""
        self.finder.clear()  # clear finder

    def focusInEvent(self, e):
        """
        Focus in event

        :param e: focus event
        """
        super(HtmlOutput, self).focusInEvent(e)
        self.window.controller.finder.focus_in(self.finder)

    def eventFilter(self, source, event):
        """
        Focus event filter

        :param source: source
        :param event: event
        """
        if event.type() == QEvent.ChildAdded and source is self and event.child().isWidgetType():
            self._detach_gl_event_filter()
            self._glwidget = event.child()
            try:
                self._glwidget.installEventFilter(self)
                self._glwidget_filter_installed = True
            except Exception:
                self._glwidget = None
                self._glwidget_filter_installed = False
        elif (event.type() == event.Type.MouseButtonPress):
            if self.tab:
                col_idx = self.tab.column_idx
                self.window.controller.ui.tabs.on_column_focus(col_idx)
        elif event.type() == QEvent.KeyPress:
            if event.key() == Qt.Key_F and (event.modifiers() & Qt.ControlModifier):
                self.find_open()
        return super().eventFilter(source, event)


class CustomWebEnginePage(QWebEnginePage):
    """Custom WebEnginePage to handle web events"""
    def __init__(self, window, parent):
        super(CustomWebEnginePage, self).__init__(window)
        self.window = window
        self.parent = parent
        self.signals = WebEnginePageSignals()
        self.findTextFinished.connect(self.on_find_finished)
        self.zoomFactorChanged.connect(self.on_view_changed)
        self.selectionChanged.connect(self.on_selection_changed)
        self.settings().setAttribute(
            QWebEngineSettings.LocalContentCanAccessFileUrls, True
        )
        self.settings().setAttribute(
            QWebEngineSettings.LocalContentCanAccessRemoteUrls, True
        )
        self.settings().setFontFamily(QWebEngineSettings.StandardFont, 'Lato')
        self.settings().setFontFamily(QWebEngineSettings.FixedFont, 'Monaspace Neon')
        self.settings().setFontFamily(QWebEngineSettings.SerifFont, 'Monaspace Neon')
        if self.window.core.config.has("zoom"):
            self.setZoomFactor(self.window.core.config.get("zoom"))

        self.loaded = False

        # bridge Python <> JavaScript
        self.bridge = Bridge(self.window)
        self.channel = QWebChannel(self)
        self.channel.registerObject("bridge", self.bridge)
        self.setWebChannel(self.channel)

    def on_find_finished(self, result):
        """
        On find text finished

        :param result: Find result
        """
        current = int(result.activeMatch())
        num = int(result.numberOfMatches())
        self.parent.finder.current_match_index = current
        self.parent.finder.matches = num
        self.parent.finder.on_find_finished()

    def set_loaded(self, loaded: bool = True):
        self.loaded = loaded

    def on_view_changed(self):
        """On view changed"""
        zoom = self.zoomFactor()
        self.window.core.config.set("zoom", zoom)
        option = self.window.controller.settings.editor.get_option('zoom')
        option['value'] = zoom
        self.window.controller.config.apply(
            parent_id='config',
            key='zoom',
            option=option,
        )

    def on_selection_changed(self):
        """On selection changed"""
        pass

    def acceptNavigationRequest(self, url,  _type, isMainFrame):
        """
        On navigation (link click) event

        :param url: URL
        :param _type: Navigation type
        :param isMainFrame: True if main frame
        """
        if _type == QWebEnginePage.NavigationTypeLinkClicked:
            self.window.core.filesystem.url.handle(url)
            return False
        return super().acceptNavigationRequest(url,  _type, isMainFrame)

    def javaScriptConsoleMessage(self, level, message, line_number, source_id):
        """
        On JavaScript console message

        :param level: log level
        :param message: message
        :param line_number: line number
        :param source_id: source ID
        """
        self.signals.js_message.emit(line_number, message, source_id)  # handled in debug controller

    def cleanup(self):
        """Cleanup method to release resources"""
        self.loaded = False
        try:
            self.findTextFinished.disconnect()
            self.zoomFactorChanged.disconnect()
            self.selectionChanged.disconnect()
        except Exception:
            pass

        if self.bridge:
            try:
                self.bridge.cleanup()
            except Exception:
                pass
            self.bridge = None

        if self.channel:
            try:
                self.channel.unregisterObject("bridge")
            except Exception:
                pass
            self.channel = None

        if self.signals:
            try:
                self.signals.deleteLater()
            except Exception:
                pass
            self.signals = None


class Bridge(QObject):
    """Bridge between Python and JavaScript"""
    def __init__(self, window):
        super(Bridge, self).__init__(window)
        self.window = window

    @Slot(str)
    def copy_text(self, text: str):
        """
        Copy text from web to clipboard

        :param text: text
        """
        self.window.controller.ctx.extra.copy_code_text(text)

    @Slot(str)
    def preview_text(self, text: str):
        """
        Preview code

        :param text: text
        """
        self.window.controller.ctx.extra.preview_code_text(text)

    @Slot(int)
    def update_scroll_position(self, pos: int):
        """
        Update scroll position from web view

        :param pos: scroll position
        """
        self.window.controller.chat.render.scroll = pos

    def cleanup(self):
        """Cleanup method to release resources"""
        if self.window:
            try:
                self.window = None
            except Exception:
                pass

        # delete the bridge object
        self.deleteLater()


class WebEngineSignals(QObject):
    save_as = Signal(str, str)
    audio_read = Signal(str)


class WebEnginePageSignals(QObject):
    js_message = Signal(int, str, str)  # on Javascript message


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2026.01.03 00:00:00                  #
# ================================================== #
from PySide6 import QtCore
from PySide6.QtCore import Qt, QObject, Signal, Slot, QEvent, QUrl, QCoreApplication, QEventLoop
from PySide6.QtWebChannel import QWebChannel
from PySide6.QtWebEngineCore import QWebEngineSettings, QWebEnginePage, QWebEngineProfile
from PySide6.QtWebEngineWidgets import QWebEngineView
from PySide6.QtGui import QAction, QIcon
from PySide6.QtWidgets import QMenu, QDialog, QVBoxLayout

from pygpt_net.core.events import RenderEvent
from pygpt_net.item.ctx import CtxMeta
from pygpt_net.core.text.web_finder import WebFinder
from pygpt_net.ui.widget.tabs.layout import FocusEventFilter
from pygpt_net.utils import trans


class ChatWebOutput(QWebEngineView):
    def __init__(self, window=None):
        """
        HTML output (WebEngine)

        :param window: Window instance
        """
        super(ChatWebOutput, self).__init__(window)
        self.window = window
        self.finder = WebFinder(window, self)
        self.loadFinished.connect(self.on_page_loaded)
        self.customContextMenuRequested.connect(self.on_context_menu)
        self.signals = WebEngineSignals(self)
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.filter = FocusEventFilter(self, self.on_focus)
        self.installEventFilter(self)
        self.plain = None
        self.html_content = None
        self.meta = None
        self.tab = None
        self.setProperty('class', 'layout-output-web')
        self.setMouseTracking(True)

        # OpenGL widgets
        self._glwidget = None
        self._glwidget_filter_installed = False
        self._unloaded = False
        self._destroyed = False

        # set the page with a shared profile
        self.setUpdatesEnabled(False)  # disable updates until the page is set, re-enable in `on_page_loaded`

        # self._profile = self._make_profile(self)
        self.setPage(CustomWebEnginePage(self.window, self, profile=None))

    def _make_profile(self, parent=None) -> QWebEngineProfile:
        """Make profile"""
        profile = QWebEngineProfile(parent)
        # self._profile.setHttpCacheType(QWebEngineProfile.MemoryHttpCache)
        # self._profile.setHttpCacheMaximumSize(32 * 1024 * 1024)  # 32MB
        profile.setPersistentCookiesPolicy(QWebEngineProfile.NoPersistentCookies)
        # self._profile.setHttpCacheType(QWebEngineProfile.NoCache)
        profile.setSpellCheckEnabled(False)
        return profile


    def _detach_gl_event_filter(self):
        """Detach OpenGL widget event filter if installed"""
        if self._glwidget and self._glwidget_filter_installed:
            try:
                self._glwidget.removeEventFilter(self)
            except Exception as e:
                self._on_delete_failed(e)
        self._glwidget = None
        self._glwidget_filter_installed = False

    def _on_delete_failed(self, e):
        """
        Handle delete failure

        :param e: Exception instance
        """
        pass
        # self.window.core.debug.log(e)

    def show_devtools(self, modal: bool = False, title: str = "DevTools"):
        """Show DevTools window"""
        if getattr(self, "_devtools_dlg", None) is None:
            dlg = QDialog(self.window)
            dlg.setWindowTitle(title)
            dlg.setModal(modal)
            dlg.setAttribute(Qt.WA_DeleteOnClose, True)
            dlg.resize(1100, 750)

            layout = QVBoxLayout(dlg)
            view = QWebEngineView(dlg)
            layout.addWidget(view)

            profile = self.page().profile()
            view.setPage(QWebEnginePage(profile, view))

            self.page().setDevToolsPage(view.page())

            dlg.destroyed.connect(lambda: setattr(self, "_devtools_dlg", None))

            self._devtools_dlg = dlg
            self._devtools_view = view

        self._devtools_dlg.show()
        self._devtools_dlg.raise_()
        self._devtools_dlg.activateWindow()

    def unload(self):
        """Unload the current page and free resources"""
        try:
            self.hide()
            p = self.page()
            p.triggerAction(QWebEnginePage.Stop)
            p.setUrl(QUrl("about:blank"))
            p.history().clear()
            p.setLifecycleState(QWebEnginePage.LifecycleState.Discarded)
        except Exception as e:
            self._on_delete_failed(e)
        finally:
            self._unloaded = True

    def on_delete(self):
        """Clean up on delete"""
        if self._destroyed:
            return
        if not self._unloaded:
            self.unload()

        self.hide()
        self._detach_gl_event_filter()

        if self.finder:
            try:
                self.finder.disconnect()
            except Exception as e:
                self._on_delete_failed(e)
            finally:
                self.finder = None

        self.tab = None
        self.meta = None

        # safely unhook signals (may not have been hooked)
        for sig, slot in (
                (self.loadFinished, self.on_page_loaded),
                (self.customContextMenuRequested, self.on_context_menu),
                (self.signals.save_as, getattr(self.window.controller.chat.render, "handle_save_as", None)),
                (self.signals.audio_read, getattr(self.window.controller.chat.render, "handle_audio_read", None)),
        ):
            if slot:
                try:
                    sig.disconnect(slot)
                except Exception as e:
                    self._on_delete_failed(e)

        page = self.page()
        page.set_loaded(False)

        try:
            page.triggerAction(QWebEnginePage.Stop)
        except Exception as e:
            self._on_delete_failed(e)

        try:
            page.setUrl(QUrl("about:blank"))
        except Exception as e:
            self._on_delete_failed(e)

        try:
            page.history().clear()
        except Exception as e:
            self._on_delete_failed(e)

        try:
            page.setLifecycleState(QWebEnginePage.LifecycleState.Discarded)
        except Exception as e:
            self._on_delete_failed(e)

        try:
            if hasattr(page, "setWebChannel"):
                page.setWebChannel(None)
        except Exception as e:
            self._on_delete_failed(e)

        prof = None
        try:
            prof = page.profile()
        except Exception as e:
            self._on_delete_failed(e)

        try:
            page.cleanup()
        except Exception as e:
            self._on_delete_failed(e)

        """
        if prof is not None:
            try:
                prof.deleteLater()
            except Exception as e:
                self._on_delete_failed(e)
        """

        try:
            self.deleteLater()
        except Exception as e:
            self._on_delete_failed(e)

        try:
            QCoreApplication.sendPostedEvents(None, QEvent.DeferredDelete)
            QCoreApplication.processEvents(QEventLoop.AllEvents, 50)
        except Exception as e:
            self._on_delete_failed(e)

        self._destroyed = True

    def eventFilter(self, source, event):
        """
        Event filter to handle child added events and mouse button presses

        :param source: QWidget - source of the event
        :param event: QEvent - event to filter
        """
        if event.type() == QEvent.ChildAdded and source is self and event.child().isWidgetType():
            self._detach_gl_event_filter()
            self._glwidget = event.child()
            try:
                self._glwidget.installEventFilter(self)
                self._glwidget_filter_installed = True
            except Exception:
                self._glwidget = None
                self._glwidget_filter_installed = False

        elif event.type() == QEvent.Type.MouseButtonPress:
            try:
                col_idx = self.tab.column_idx
                self.window.controller.ui.tabs.on_column_focus(col_idx)
            except Exception:
                pass

        return super().eventFilter(source, event)

    def on_focus(self, widget):
        """
        On widget clicked

        :param widget: QWidget - widget that received focus
        """
        if self.tab is not None:
            self.window.controller.ui.tabs.on_column_focus(self.tab.column_idx)
        self.setFocus()

    def set_tab(self, tab):
        """
        Set tab for this output

        :param tab: Tab instance
        """
        self.tab = tab

    def get_tab(self):
        """
        Get current tab

        :return: Tab instance
        """
        return self.tab

    def set_meta(self, meta: CtxMeta):
        """
        Set meta for this output

        :param meta: CtxMeta instance
        """
        self.meta = meta

    def set_plaintext(self, text: str):
        """
        Set plain text content

        :param text: str - plain text content
        """
        self.plain = text

    def set_html_content(self, html: str):
        """
        Set HTML content

        :param html: str - HTML content
        """
        self.html_content = "<html>" + html + "</html>"

    def on_context_menu(self, position):
        """
        Context menu event

        :param position: QPoint - position of the context menu
        """
        menu = QMenu(self)

        has_selection = self.page().hasSelection()

        if has_selection:
            # copy
            action = QAction(QIcon(":/icons/copy.svg"), trans('action.copy'), self)
            action.triggered.connect(self.copy_selected_text)
            menu.addAction(action)

            # audio read (get text only on click, don't copy immediately)
            action = QAction(QIcon(":/icons/volume.svg"), trans('text.context_menu.audio.read'), self)
            action.triggered.connect(self._read_selected_text)
            menu.addAction(action)

            # copy to
            copy_to_menu = self.window.ui.context_menu.get_copy_to_menu(menu, self.get_selected_text())
            menu.addMenu(copy_to_menu)

            # save as (selected) - get selection at the moment of click
            action = QAction(QIcon(":/icons/save.svg"), trans('action.save_selection_as'), self)
            action.triggered.connect(self._save_selected_txt)
            menu.addAction(action)
        else:
            # select all
            action = QAction(QIcon(":/icons/copy.svg"), trans('action.select_all'), self)
            action.triggered.connect(self.select_all_text)
            menu.addAction(action)

            # save as (all) - plain (lazy normalization only on click)
            action = QAction(QIcon(":/icons/save.svg"), trans('action.save_as') + " (text)", self)
            action.triggered.connect(self._save_as_text)
            menu.addAction(action)

            # save as (all) - html
            action = QAction(QIcon(":/icons/save.svg"), trans('action.save_as') + " (html)", self)
            action.triggered.connect(self._save_as_html)
            menu.addAction(action)

        # Add zoom submenu
        zoom_menu = self.window.ui.context_menu.get_zoom_menu(self, "zoom", self.get_zoom_value(), self.on_zoom_changed)
        menu.addMenu(zoom_menu)

        action = QAction(QIcon(":/icons/search.svg"), trans('text.context_menu.find'), self)
        action.triggered.connect(self.find_open)
        menu.addAction(action)

        menu.exec_(self.mapToGlobal(position))

    def on_zoom_changed(self, zoom: float):
        """
        On zoom changed from context menu

        :param zoom: float - new zoom factor
        """
        p = self.page()
        if p:
            self.window.core.config.set("zoom", zoom)
            self.update_zoom()

    @Slot()
    def _save_selected_txt(self):
        """Save selected content as text file"""
        self.signals.save_as.emit(self.get_selected_text(), 'txt')

    @Slot()
    def _read_selected_text(self):
        """
        Read selected text using text-to-speech
        """
        selected_text = self.get_selected_text()
        if selected_text:
            self.signals.audio_read.emit(selected_text)

    @Slot()
    def _save_as_text(self):
        """
        Save current content as text file
        """
        # TODO: normalize text (remove extra spaces, newlines, etc.)
        self.page().toPlainText(lambda txt: self.signals.save_as.emit(txt, 'txt'))

    @Slot()
    def _save_as_html(self):
        """
        Save current content as HTML file
        """
        self.page().toHtml(lambda html: self.signals.save_as.emit(html, 'html'))

    def update_zoom(self):
        """Update zoom from config"""
        try:
            if self.window.core.config.has("zoom"):
                p = self.page()
                if p:
                    p.setZoomFactor(self.window.core.config.get("zoom"))
        except Exception:
            pass

    def on_focus_js(self):
        """Focus JavaScript"""
        if self.tab is not None:
            self.window.controller.ui.tabs.on_column_focus(self.tab.column_idx)

    def get_zoom_value(self) -> float:
        """
        Get current zoom value

        :return: zoom factor
        """
        p = self.page()
        return p.zoomFactor() if p else 1.0

    def reset_current_content(self):
        """Reset current content"""
        self.plain = ""
        self.html_content = ""

    def on_page_loaded(self, success):
        """
        Page loaded event handler

        :param success: bool - True if page loaded successfully, False otherwise
        """
        if success:
            event = RenderEvent(RenderEvent.ON_PAGE_LOAD, {
                "meta": self.meta,
                "tab": self.tab,
            })
            self.window.dispatch(event)
            self.setUpdatesEnabled(True)
            p = self.page()
            if p and isinstance(p, CustomWebEnginePage):
                p.set_loaded(True)

    def get_selected_text(self) -> str:
        p = self.page()
        return p.selectedText() if p else ""

    def copy_selected_text(self):
        p = self.page()
        if p:
            p.triggerAction(QWebEnginePage.Copy)

    def select_all_text(self):
        p = self.page()
        if p:
            p.triggerAction(QWebEnginePage.SelectAll)

    def unselect_text(self):
        p = self.page()
        if p:
            p.triggerAction(QWebEnginePage.Unselect)

    def find_open(self):
        self.window.controller.finder.open(self.finder)

    def on_update(self):
        if self.finder:
            self.finder.clear()

    def focusInEvent(self, e):
        super(ChatWebOutput, self).focusInEvent(e)
        self.window.controller.finder.focus_in(self.finder)


class CustomWebEnginePage(QWebEnginePage):
    """Custom WebEnginePage to handle web events"""

    def __init__(self, window, view, profile: QWebEngineProfile = None):

        # use the profile if provided, otherwise the default
        if profile is not None:
            super(CustomWebEnginePage, self).__init__(profile, view)
        else:
            super(CustomWebEnginePage, self).__init__(view)

        self.window = window
        self.view = view
        self.loaded = False

        # signals have parent=page (automatic cleanup)
        self.signals = WebEnginePageSignals(self)

        self.findTextFinished.connect(self.on_find_finished)
        self.zoomFactorChanged.connect(self.on_view_changed)

        s = self.settings()
        s.setAttribute(QWebEngineSettings.LocalContentCanAccessFileUrls, True)
        s.setAttribute(QWebEngineSettings.LocalContentCanAccessRemoteUrls, True)
        s.setFontFamily(QWebEngineSettings.StandardFont, 'Lato')
        s.setFontFamily(QWebEngineSettings.FixedFont, 'Monaspace Neon')
        s.setFontFamily(QWebEngineSettings.SerifFont, 'Monaspace Neon')
        s.setAttribute(QWebEngineSettings.PluginsEnabled, False)
        s.setAttribute(QWebEngineSettings.FullScreenSupportEnabled, False)
        s.setAttribute(QWebEngineSettings.ScreenCaptureEnabled, False)
        s.setAttribute(QWebEngineSettings.WebGLEnabled, False)
        s.setAttribute(QWebEngineSettings.Accelerated2dCanvasEnabled, False)
        s.setAttribute(QWebEngineSettings.ErrorPageEnabled, False)

        if self.window.core.config.has("zoom"):
            self.setZoomFactor(self.window.core.config.get("zoom"))

        # bridge Python <> JavaScript
        # KEY: parent=page, -> disappears along with the page
        self.bridge = Bridge(self.window, parent=self)
        self.channel = QWebChannel(self)
        self.channel.registerObject("bridge", self.bridge)
        self.setWebChannel(self.channel)

    def set_loaded(self, loaded: bool = True):
        self.loaded = loaded

    def on_find_finished(self, result):
        current = int(result.activeMatch())
        num = int(result.numberOfMatches())
        self.view.finder.current_match_index = current
        self.view.finder.matches = num
        self.view.finder.on_find_finished()

    def on_view_changed(self):
        if not self.loaded:
            return
        zoom = self.zoomFactor()
        self.window.core.config.set("zoom", zoom)
        option = self.window.controller.settings.editor.get_option('zoom')
        option['value'] = zoom
        self.window.controller.config.apply(
            parent_id='config',
            key='zoom',
            option=option,
        )

    def acceptNavigationRequest(self, url, _type, isMainFrame):
        if _type == QWebEnginePage.NavigationTypeLinkClicked:
            self.window.core.filesystem.url.handle(url)
            return False
        return super().acceptNavigationRequest(url, _type, isMainFrame)

    def javaScriptConsoleMessage(self, level, message, line_number, source_id):
        print("[JS CONSOLE] Line", line_number, ":", message)
        self.signals.js_message.emit(line_number, message, source_id)  # handled in debug controller

    def cleanup(self):
        """Cleanup method to release resources"""
        self.loaded = False
        try:
            self.findTextFinished.disconnect()
            self.zoomFactorChanged.disconnect()
            self.selectionChanged.disconnect()
        except Exception:
            pass

        if self.bridge:
            try:
                self.bridge.cleanup()
            except Exception:
                pass
            self.bridge = None

        if self.channel:
            try:
                self.channel.unregisterObject("bridge")
            except Exception:
                pass
            self.channel = None

        if self.signals:
            try:
                self.signals.deleteLater()
            except Exception:
                pass
            self.signals = None


class Bridge(QObject):
    """Bridge between Python and JavaScript"""

    def __init__(self, window, parent=None):
        super(Bridge, self).__init__(parent)
        self.window = window

    chunk = Signal(str, str, str)  # name, chunk, type
    node = Signal(str)  # JSON payload
    nodeReplace = Signal(str)  # JSON payload
    nodeInput = Signal(str)  # raw text
    readyChanged = Signal(bool)

    @Slot(int)
    def js_ready(self, pid: int):
        self.readyChanged.emit(True)
        self.window.controller.chat.render.on_js_ready(pid)

    @Slot(str)
    def log(self, text: str):
        print(f"JS log: {text}")

    @Slot(str)
    def copy_text(self, text: str):
        self.window.controller.ctx.extra.copy_code_text(text)

    @Slot(str)
    def preview_text(self, text: str):
        self.window.controller.ctx.extra.preview_code_text(text)

    @Slot(str)
    def run_text(self, text: str):
        self.window.controller.ctx.extra.run_code_text(text)

    @Slot(int)
    def update_scroll_position(self, pos: int):
        self.window.controller.chat.render.scroll = pos

    def cleanup(self):
        """Cleanup method to release resources"""
        if self.window:
            try:
                self.window = None
            except Exception:
                pass

        # delete the bridge object
        self.deleteLater()


class WebEngineSignals(QObject):
    save_as = Signal(str, str)
    audio_read = Signal(str)


class WebEnginePageSignals(QObject):
    js_message = Signal(int, str, str)  # on JavaScript message


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.08.15 23:00:00                  #
# ================================================== #

from typing import Optional, List

from PySide6.QtCore import Slot, QTimer

from pygpt_net.core.events import RenderEvent
from pygpt_net.core.render.base import BaseRenderer
from pygpt_net.core.render.markdown.renderer import Renderer as MarkdownRenderer
from pygpt_net.core.render.plain.renderer import Renderer as PlainTextRenderer
from pygpt_net.core.render.web.renderer import Renderer as WebRenderer
from pygpt_net.core.tabs.tab import Tab
from pygpt_net.core.text.utils import output_html2text, output_clean_html
from pygpt_net.item.ctx import CtxItem, CtxMeta


class Render:

    _STATE_EVENTS = (
        RenderEvent.STATE_IDLE,
        RenderEvent.STATE_BUSY,
        RenderEvent.STATE_ERROR,
    )

    def __init__(self, window=None):
        """
        Render controller

        :param window: Window instance
        """
        self.window = window
        self.markdown_renderer = MarkdownRenderer(window)
        self.plaintext_renderer = PlainTextRenderer(window)
        self.web_renderer = WebRenderer(window)
        self.engine = None
        self.scroll = 0
        self.renderer = None

    def setup(self) -> None:
        """Setup render"""
        self.engine = self.window.core.config.get("render.engine")
        if self.window.core.config.get('render.plain'):
            self.renderer = self.plaintext_renderer
        else:
            self.renderer = self.web_renderer if self.engine == "web" else self.markdown_renderer

    def prepare(self) -> None:
        """Prepare render"""
        self.markdown_renderer.prepare()
        self.plaintext_renderer.prepare()
        self.web_renderer.prepare()

    def handle(self, event: RenderEvent) -> None:
        """
        Handle render event

        :param event: RenderEvent
        """
        name = event.name
        data = event.data or {}

        if name == RenderEvent.BEGIN:
            self.begin(data.get("meta"), data.get("ctx"), data.get("stream", False))
        elif name == RenderEvent.END:
            self.end(data.get("meta"), data.get("ctx"), data.get("stream", False))
        elif name == RenderEvent.RELOAD:
            self.reload()
        elif name == RenderEvent.RESET:
            self.reset(data.get("meta"))
        elif name == RenderEvent.PREPARE:
            self.prepare()

        elif name == RenderEvent.STREAM_BEGIN:
            self.stream_begin(data.get("meta"), data.get("ctx"))
        elif name == RenderEvent.STREAM_APPEND:
            self.instance().append_chunk(
                data.get("meta"),
                data.get("ctx"),
                data.get("chunk", ""),
                data.get("begin", False),
            )
        elif name == RenderEvent.STREAM_NEXT:
            self.next_chunk(data.get("meta"), data.get("ctx"))
        elif name == RenderEvent.STREAM_END:
            self.stream_end(data.get("meta"), data.get("ctx"))

        elif name == RenderEvent.ON_PAGE_LOAD:
            self.on_page_loaded(data.get("meta"), data.get("tab"))
        elif name == RenderEvent.ON_THEME_CHANGE:
            self.on_theme_change()
        elif name == RenderEvent.ON_LOAD:
            self.on_load(data.get("meta"))
        elif name == RenderEvent.FRESH:
            self.fresh(data.get("meta"))
        elif name == RenderEvent.ON_TS_ENABLE:
            self.on_enable_timestamp(live=data.get("initialized", False))
        elif name == RenderEvent.ON_TS_DISABLE:
            self.on_disable_timestamp(live=data.get("initialized", False))
        elif name == RenderEvent.ON_EDIT_ENABLE:
            self.on_enable_edit(live=data.get("initialized", False))
        elif name == RenderEvent.ON_EDIT_DISABLE:
            self.on_disable_edit(live=data.get("initialized", False))
        elif name == RenderEvent.ON_SWITCH:
            self.switch()

        elif name == RenderEvent.CLEAR_INPUT:
            self.clear_input()
        elif name == RenderEvent.CLEAR_OUTPUT:
            self.clear_output(data.get("meta"))
        elif name == RenderEvent.CLEAR_ALL:
            self.clear_all()
        elif name == RenderEvent.CLEAR:
            self.clear(data.get("meta"))

        elif name == RenderEvent.TOOL_UPDATE:
            self.tool_output_update(data.get("meta"), data.get("tool_data"))
        elif name == RenderEvent.TOOL_CLEAR:
            self.tool_output_clear(data.get("meta"))
        elif name == RenderEvent.TOOL_BEGIN:
            self.tool_output_begin(data.get("meta"))
        elif name == RenderEvent.TOOL_END:
            self.tool_output_end()

        elif name == RenderEvent.CTX_APPEND:
            self.append_context(data.get("meta"), data.get("items"), data.get("clear", True))
        elif name == RenderEvent.INPUT_APPEND:
            self.append_input(
                data.get("meta"),
                data.get("ctx"),
                data.get("flush", True),
                data.get("append", False),
            )
        elif name == RenderEvent.OUTPUT_APPEND:
            self.append_output(data.get("meta"), data.get("ctx"))

        elif name == RenderEvent.EXTRA_APPEND:
            self.append_extra(data.get("meta"), data.get("ctx"), data.get("footer", False))
        elif name == RenderEvent.EXTRA_END:
            self.end_extra(data.get("meta"), data.get("ctx"))

        elif name == RenderEvent.LIVE_APPEND:
            self.append_live(
                data.get("meta"),
                data.get("ctx"),
                data.get("chunk", ""),
                data.get("begin", False),
            )
        elif name == RenderEvent.LIVE_CLEAR:
            self.clear_live(data.get("meta"), data.get("ctx"))

        elif name == RenderEvent.ACTION_REGEN_SUBMIT:
            self.on_reply_submit(data.get("ctx"))
        elif name == RenderEvent.ACTION_EDIT_SUBMIT:
            self.on_edit_submit(data.get("ctx"))

        elif name in self._STATE_EVENTS:
            meta = data.get("meta") or self.window.core.ctx.get_current_meta()
            self.on_state_changed(name, meta)

        elif name == RenderEvent.ITEM_DELETE_ID:
            self.remove_item(data.get("ctx"))
        elif name == RenderEvent.ITEM_DELETE_FROM_ID:
            self.remove_items_from(data.get("ctx"))

    def on_state_changed(self, state: str, meta: Optional[CtxMeta] = None) -> None:
        """
        Handle state change event

        :param state: State name
        :param meta: Context meta
        """
        self.instance().state_changed(state, meta)

    def append_live(self, meta: CtxMeta, ctx: CtxItem, text_chunk: str, begin: bool = False) -> None:
        """
        Append live text chunk to output

        :param meta: context meta
        :param ctx: context item
        :param text_chunk: text chunk
        :param begin: if it is the beginning of the stream
        """
        self.instance().append_live(meta, ctx, text_chunk, begin)

    def clear_live(self, meta: CtxMeta, ctx: CtxItem) -> None:
        """
        Clear live output

        :param meta: context meta
        :param ctx: context item
        """
        self.instance().clear_live(meta, ctx)
        self.update()

    def get_pid(self, meta: CtxMeta) -> int:
        """
        Get PID for context meta

        :param meta: context meta
        :return: PID
        """
        return self.instance().get_pid(meta)

    def begin(self, meta: CtxMeta, ctx: CtxItem, stream: bool = False) -> None:
        """
        Render begin

        :param meta: context meta
        :param ctx: context item
        :param stream: True if it is a stream
        """
        self.instance().begin(meta, ctx, stream)
        self.update()

    def end(self, meta: CtxMeta, ctx: CtxItem, stream: bool = False) -> None:
        """
        Render end

        :param meta: context meta
        :param ctx: context item
        :param stream: True if it is a stream
        """
        self.instance().end(meta, ctx, stream)
        self.update()

    def end_extra(self, meta: CtxMeta, ctx: CtxItem, stream: bool = False) -> None:
        """
        Render end extra

        :param meta: context meta
        :param ctx: context item
        :param stream: True if it is a stream
        """
        self.instance().end_extra(meta, ctx, stream)
        self.update()

    def stream_begin(self, meta: CtxMeta, ctx: CtxItem) -> None:
        """
        Render stream begin

        :param meta: context meta
        :param ctx: context item
        """
        self.instance().stream_begin(meta, ctx)
        self.update()

    def stream_end(self, meta: CtxMeta, ctx: CtxItem) -> None:
        """
        Render stream end

        :param meta: context meta
        :param ctx: context item
        """
        self.instance().stream_end(meta, ctx)
        self.update()

    def clear_output(self, meta: Optional[CtxMeta] = None) -> None:
        """
        Clear current active output

        :param meta: context meta
        """
        self.instance().clear_output(meta) # TODO: get meta id on load
        self.update()

    def clear_input(self) -> None:
        """Clear input"""
        self.instance().clear_input()

    def on_load(self, meta: Optional[CtxMeta] = None) -> None:
        """
        On load (meta)

        :param meta: context meta
        """
        self.instance().on_load(meta)
        self.update()
        self.window.controller.ui.tabs.update_tooltip(meta.name)  # update tab tooltip

    def fresh(self, meta: Optional[CtxMeta] = None) -> None:
        """
        On load (meta)

        :param meta: context meta
        """
        self.instance().fresh(meta)

    def reset(self, meta: Optional[CtxMeta] = None) -> None:
        """
        Reset current meta

        :param meta: Context meta
        """
        self.instance().reset(meta)  # TODO: get meta id on load
        self.update()

    def reload(self) -> None:
        """Reload current output"""
        self.instance().reload()  # TODO: or all outputs?
        self.update()

    def append_context(self, meta: CtxMeta, items: List[CtxItem], clear: bool = True) -> None:
        """
        Append all context to output

        :param meta: context meta
        :param items: context items
        :param clear: True if clear all output before append
        """
        self.instance().append_context(meta, items, clear)
        self.update()

    def append_input(self, meta: CtxMeta, ctx: CtxItem, flush: bool = True, append: bool = False) -> None:
        """
        Append text input to output

        :param meta: context meta
        :param ctx: context item
        :param flush: True if flush output
        :param append: True to force append node
        """
        self.instance().append_input(meta, ctx, flush=flush, append=append)
        self.update()

    def append_output(self, meta: CtxMeta, ctx: CtxItem) -> None:
        """
        Append text output to output

        :param meta: context meta
        :param ctx: context item
        """
        self.instance().append_output(meta, ctx)
        self.update()

    def append_extra(self, meta: CtxMeta, ctx: CtxItem, footer: bool = False) -> None:
        """
        Append extra data (images, files, etc.) to output

        :param meta: context meta
        :param ctx: context item
        :param footer: True if it is a footer
        """
        self.instance().append_extra(meta, ctx, footer)
        self.update()

    def append_chunk(self, meta: CtxMeta, ctx: CtxItem, text_chunk: str, begin: bool = False) -> None:
        """
        Append output stream chunk to output

        :param meta: context meta
        :param ctx: context item
        :param text_chunk: text chunk
        :param begin: if it is the beginning of the stream
        """
        self.instance().append_chunk(meta, ctx, text_chunk, begin)

    def next_chunk(self, meta: CtxMeta, ctx: CtxItem) -> None:
        """
        Flush current stream and start with new chunks

        :param meta: context meta
        :param ctx: context item
        """
        self.instance().next_chunk(meta, ctx)
        self.update()

    def on_enable_edit(self, live: bool = True) -> None:
        """
        On enable edit icons - global

        :param live: True if live update
        """
        self.instance().on_enable_edit(live)
        self.update()

    def on_disable_edit(self, live: bool = True) -> None:
        """
        On disable edit icons - global

        :param live: True if live update
        """
        self.instance().on_disable_edit(live)
        self.update()

    def on_enable_timestamp(self, live: bool = True) -> None:
        """
        On enable timestamp - global

        :param live: True if live update
        """
        self.instance().on_enable_timestamp(live)
        self.update()

    def on_disable_timestamp(self, live: bool = True) -> None:
        """
        On disable timestamp - global

        :param live: True if live update
        """
        self.instance().on_disable_timestamp(live)
        self.update()

    def remove_item(self, ctx: CtxItem) -> None:
        """
        Remove item from output

        :param ctx: context item
        """
        self.instance().remove_item(ctx)
        self.update()

    def remove_items_from(self, ctx: CtxItem) -> None:
        """
        Remove item from output

        :param ctx: context item
        """
        self.instance().remove_items_from(ctx)
        self.update()

    def on_edit_submit(self, ctx: CtxItem) -> None:
        """
        On edit submit

        :param ctx: context item
        """
        self.instance().on_edit_submit(ctx)
        self.update()

    def on_remove_submit(self, ctx: CtxItem) -> None:
        """
        On remove submit

        :param ctx: context item
        """
        self.instance().on_remove_submit(ctx)
        self.update()

    def on_reply_submit(self, ctx: CtxItem) -> None:
        """
        On regenerate submit

        :param ctx: context item
        """
        self.instance().on_reply_submit(ctx)
        self.update()

    def on_page_loaded(self, meta: CtxMeta, tab: Optional[Tab] = None) -> None:
        """
        On page loaded callback

        :param meta: context meta
        :param tab: Tab
        """
        self.instance().on_page_loaded(meta, tab)
        self.update()

    def on_theme_change(self) -> None:
        """On theme change - global"""
        if self.get_engine() == "web":
            self.web_renderer.on_theme_change()
        elif self.get_engine() == "legacy":
            self.markdown_renderer.on_theme_change()
        self.update()

    def get_scroll_position(self) -> int:
        """
        Get scroll position - active

        :return: scroll position
        """
        return self.instance().get_scroll_position()

    def set_scroll_position(self, position: int) -> None:
        """
        Set scroll position - active

        :param position: scroll position
        """
        self.instance().set_scroll_position(position)
        self.update()

    def update(self) -> None:
        """On update - active"""
        nodes = self.window.ui.nodes
        outputs = nodes.get('output')
        if not outputs:
            return
        for node in outputs.values():
            node.on_update()

    def clear(self, meta: CtxMeta) -> None:
        """
        Clear renderer

        :param meta: ctx meta instance
        """
        inst = self.instance()
        inst.reset(meta)
        inst.clear_output(meta)
        self.update()

    def clear_all(self) -> None:
        """Clear all"""
        self.instance().clear_all()
        self.update()

    def remove_pid(self, pid: int) -> None:
        """
        Remove PID from renderer

        :param pid: PID to remove
        """
        self.plaintext_renderer.remove_pid(pid)
        self.markdown_renderer.remove_pid(pid)
        self.web_renderer.remove_pid(pid)

    def tool_output_append(self, meta: CtxMeta, content: str) -> None:
        """
        Add tool output (append)

        :param meta: context meta
        :param content: content
        """
        self.instance().tool_output_append(meta, content)
        self.update()

    def tool_output_update(self, meta: CtxMeta, content: str) -> None:
        """
        Replace tool output

        :param meta: context meta
        :param content: content
        """
        self.instance().tool_output_update(meta, content)
        self.update()

    def tool_output_clear(self, meta: CtxMeta) -> None:
        """
        Clear tool output

        :param meta: context meta
        """
        self.instance().tool_output_clear(meta)
        self.update()

    def tool_output_begin(self, meta: CtxMeta) -> None:
        """
        Begin tool output

        :param meta: context meta
        """
        self.instance().tool_output_begin(meta)
        self.update()

    def tool_output_end(self) -> None:
        """
        End tool output

        :param meta: context meta
        """
        self.instance().tool_output_end()
        self.update()

    def on_js_ready(self, pid: int) -> None:
        """
        On JS ready - called from WebEngine when Bridge is ready

        Web renderer only.

        :param pid: PID
        """
        if self.get_engine() == "web":
            self.web_renderer.on_js_ready(pid)

    def get_engine(self) -> str:
        """
        Get current render engine ID

        :return: engine name
        """
        return self.engine

    def switch(self) -> None:
        """
        Switch renderer (markdown/web <==> plain text) - active, TODO: remove from settings, leave only checkbox
        """
        plain = self.window.core.config.get('render.plain')
        nodes = self.window.ui.nodes
        if plain:
            self.window.controller.theme.markdown.clear()
            nodes['output.timestamp'].setVisible(True)
            outputs = nodes.get('output', {})
            outputs_plain = nodes.get('output_plain', {})
            for pid, w_plain in outputs_plain.items():
                try:
                    w = outputs.get(pid)
                    if w and w_plain:
                        w.setVisible(False)
                        w_plain.setVisible(True)
                except Exception:
                    continue
        else:
            nodes['output.timestamp'].setVisible(False)
            self.window.controller.theme.markdown.update(force=True)
            outputs = nodes.get('output', {})
            outputs_plain = nodes.get('output_plain', {})
            for pid, w in outputs.items():
                try:
                    w_plain = outputs_plain.get(pid)
                    if w and w_plain:
                        w.setVisible(True)
                        w_plain.setVisible(False)
                except Exception:
                    continue

        if plain:
            self.renderer = self.plaintext_renderer
        else:
            if self.engine == "web":
                self.renderer = self.web_renderer
            else:
                self.renderer = self.markdown_renderer

        self.window.controller.ctx.refresh()

    def instance(self) -> BaseRenderer:
        """
        Get instance of current renderer

        :return: Renderer instance
        """
        if self.renderer:
            return self.renderer
        if self.window.core.config.get('render.plain'):
            return self.plaintext_renderer
        else:
            if self.engine == "web":
                return self.web_renderer
            else:
                return self.markdown_renderer

    @Slot(str, str)
    def handle_save_as(self, text: str, type: str = 'txt') -> None:
        """
        Handle save as signal  # TODO: move to another class

        :param text: Data to save
        :param type: File type
        """
        if type == 'html':
            text = output_clean_html(text)
        else:
            text = output_html2text(text)
        # fix: QTimer required here to prevent crash if signal emitted from WebEngine window
        QTimer.singleShot(0, lambda: self.window.controller.chat.common.save_text(text, type))

    @Slot(str)
    def handle_audio_read(self, text: str):
        """
        Handle audio read signal

        :param text: Text to read  # TODO: move to another class
        """
        self.window.controller.audio.read_text(text)

    def get_pid_data(self, pid: int) -> Optional[CtxItem]:
        """
        Get PID data

        :param pid: PID
        :return: CtxItem or None
        """
        return self.instance().get_pid_data(pid)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.08.31 23:00:00                  #
# ================================================== #

from typing import Any, Optional

from pygpt_net.core.bridge import BridgeContext
from pygpt_net.core.types import (
    MODE_ASSISTANT,
    MODE_IMAGE,
    MODE_AUDIO,
)
from pygpt_net.core.events import Event, AppEvent, RenderEvent, KernelEvent
from pygpt_net.item.ctx import CtxItem


class Output:

    STATE_PARAMS = {
        "id": "chat",
    }

    NOT_STREAM_MODES = (
        MODE_ASSISTANT,
        MODE_IMAGE
    )

    def __init__(self, window=None):
        """
        Output controller

        :param window: Window instance
        """
        self.window = window

    def handle(
            self,
            ctx: CtxItem,
            mode: str,
            stream: bool = False,
            is_response: bool = False,
            reply: bool = False,
            internal: bool = False,
            context: Optional[BridgeContext] = None,
            extra: Optional[dict] = None
    ):
        """
        Handle response from LLM

        :param ctx: CtxItem
        :param mode: mode (global)
        :param stream: stream enabled (local)
        :param is_response: Is response output
        :param reply: is reply
        :param internal: is internal command
        :param context: BridgeContext (optional)
        :param extra: Extra data (optional)
        """
        self.window.dispatch(KernelEvent(KernelEvent.STATE_BUSY))  # state: busy

        # if stream then append chunk by chunk
        end = True
        if stream:  # local, not global config
            if mode not in self.NOT_STREAM_MODES:
                end = False  # don't end if stream mode, append chunk by chunk
                self.window.controller.chat.stream.append(
                    ctx=ctx,
                    mode=mode,
                    is_response=is_response,
                    reply=reply,
                    internal=internal,
                    context=context,
                    extra=extra,
                )

        if end:
            self.handle_after(
                ctx=ctx,
                mode=mode,
                stream=stream,
            )

    def handle_after(
            self,
            ctx: CtxItem,
            mode: str,
            stream: bool = False
    ):
        """
        Handle response from LLM

        :param ctx: CtxItem
        :param mode: mode (global)
        :param stream: stream enabled (local)
        """
        core = self.window.core
        dispatch = self.window.dispatch
        log = self.window.controller.chat.log

        # check if tool calls detected
        if ctx.tool_calls:
            # if not internal commands in a text body then append tool calls as commands (prevent double commands)
            if not core.command.has_cmds(ctx.output):
                core.command.append_tool_calls(ctx)  # append tool calls as commands
                if not isinstance(ctx.extra, dict):
                    ctx.extra = {}
                ctx.extra["tool_calls"] = ctx.tool_calls
                stream = False  # disable stream mode, show tool calls at the end
                log("Tool call received...")
            else:  # prevent execute twice
                log("Ignoring tool call because command received...")

        # event: context after
        dispatch(Event(Event.CTX_AFTER, {
            'mode': mode,
        }, ctx=ctx))

        log("Appending output to chat window...")

        # only append output if not in stream mode, TODO: plugin output add
        stream_global = core.config.get('stream', False)
        if not stream:
            if stream_global:  # use global stream settings here to persist previously added input
                dispatch(RenderEvent(RenderEvent.INPUT_APPEND, {
                    "meta": ctx.meta,
                    "ctx": ctx,
                    "flush": True,
                    "append": True,
                }))

            dispatch(RenderEvent(RenderEvent.OUTPUT_APPEND, {
                "meta": ctx.meta,
                "ctx": ctx,
            }))
            dispatch(RenderEvent(RenderEvent.EXTRA_APPEND, {
                "meta": ctx.meta,
                "ctx": ctx,
                "footer": True,
            }))  # + icons

        self.handle_complete(ctx)

    def handle_complete(self, ctx: CtxItem):
        """
        Handle completed context

        :param ctx: CtxItem
        """
        core = self.window.core
        controller = self.window.controller
        dispatch = self.window.dispatch
        mode = core.config.get('mode')

        # post update context, store last mode, etc.
        core.ctx.post_update(mode)
        core.ctx.store()

        controller.ctx.update_ctx()
        controller.ctx.store_history(ctx, "output")  # store to history

        controller.chat.audio.handle_output(ctx)  # handle audio output
        controller.chat.common.auto_unlock(ctx)  # unlock input if allowed
        if mode != MODE_AUDIO:
            controller.chat.common.show_response_tokens(ctx)  # update tokens
            dispatch(KernelEvent(KernelEvent.STATE_IDLE, self.STATE_PARAMS))  # state: idle
        else:
            if not controller.audio.is_recording():
                self.window.update_status("...")  # wait for audio

    def post_handle(
            self,
            ctx: CtxItem,
            mode: str,
            stream: bool = False,
            reply: bool = False,
            internal: bool = False
    ):
        """
        Post handle results

        :param ctx: CtxItem
        :param mode: mode (global)
        :param stream: stream mode
        :param reply: is reply
        :param internal: is internal
        """
        core = self.window.core
        controller = self.window.controller
        dispatch = self.window.dispatch

        # if commands enabled: post-execute commands (not assistant mode)
        if mode != MODE_ASSISTANT:
            ctx.clear_reply()  # reset results
            expert_calls = controller.agent.experts.handle(ctx)
            if expert_calls == 0:  # handle commands only if no expert calls in queue
                controller.chat.command.handle(ctx)

            ctx.from_previous()  # append previous result again before save
            core.ctx.update_item(ctx)  # update ctx in DB

        # render: end
        if ctx.sub_calls == 0:  # if no experts called
            dispatch(RenderEvent(RenderEvent.END, {
                "meta": ctx.meta,
                "ctx": ctx,
                "stream": stream,
            }))

        controller.chat.common.auto_unlock(ctx)  # unlock input if allowed
        controller.ctx.prepare_summary(ctx)  # prepare ctx name

        if self.window.state != self.window.STATE_ERROR and mode != MODE_ASSISTANT:
            dispatch(KernelEvent(KernelEvent.STATE_IDLE, self.STATE_PARAMS))  # state: idle

    def handle_end(
            self,
            ctx: CtxItem,
            mode: str
    ):
        """
        Handle context end (finish output)

        :param ctx: CtxItem
        :param mode: mode
        """
        controller = self.window.controller
        dispatch = self.window.dispatch
        log = controller.chat.log

        controller.attachment.cleanup(ctx)  # clear after send
        log(f"Context: END: {ctx}" if self.window.core.config.get("log.ctx") else "Context: END.")

        # event: context end
        dispatch(Event(Event.CTX_END, {
            'mode': mode,
        }, ctx=ctx))
        controller.chat.input.generating = False  # unlock

        log("End.")
        dispatch(AppEvent(AppEvent.CTX_END))  # app event

        # restore state to idle if no errors
        if self.window.state != self.window.STATE_ERROR:
            dispatch(KernelEvent(KernelEvent.STATE_IDLE, self.STATE_PARAMS))

        if mode != MODE_ASSISTANT:
            controller.kernel.stack.handle()  # handle reply
            dispatch(RenderEvent(RenderEvent.RELOAD))  # reload chat window


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.08.31 23:00:00                  #
# ================================================== #

import numpy as np

from PySide6.QtCore import Qt
from PySide6.QtMultimedia import QAudioFormat, QAudioSink
from PySide6.QtCore import QTimer, QObject


class RealtimeSession(QObject):
    """Global realtime session: pumps PCM bytes to QAudioSink without blocking the GUI."""
    def __init__(
            self,
            device,
            fmt: QAudioFormat,
            parent=None,
            volume_emitter: callable = None
    ):
        """
        Initialize the session.

        :param device: QAudioDeviceInfo
        :param fmt: QAudioFormat
        :param parent: QObject
        :param volume_emitter: optional callable to emit volume level (0-100)
        """
        super().__init__(parent)
        self._device = device
        self.format = fmt

        # NOTE: very simple sink with modest HW buffer (~300 ms) to avoid underruns
        self.sink = QAudioSink(device, fmt)
        hw_ms = 300
        bs = int(max(fmt.sampleRate() * fmt.channelCount() * fmt.bytesPerSample() * (hw_ms / 1000.0), 8192))
        self.sink.setBufferSize(bs)

        self.io = self.sink.start()
        if self.io is None:
            raise RuntimeError("QAudioSink.start() returned None (no IO for writing)")

        # user buffer and simple timing
        self.buffer = bytearray()
        self.final = False

        # format helpers
        self.frame_bytes = max(1, fmt.channelCount() * fmt.bytesPerSample())
        self.bytes_per_ms = max(1, int(fmt.sampleRate() * fmt.channelCount() * fmt.bytesPerSample() / 1000))

        # NOTE: keep writes reasonably sized
        self.min_write_bytes = max(self.bytes_per_ms * 20, self.frame_bytes)  # ~20 ms
        self.max_write_bytes = max(self.bytes_per_ms * 100, self.frame_bytes) # ~100 ms

        # small tail of silence to avoid end clicks
        self.tail_ms = 60  # ~60 ms

        # very small pump
        self.timer = QTimer(self)
        self.timer.setTimerType(Qt.PreciseTimer)
        self.timer.setInterval(10)  # ~10 ms
        self.timer.timeout.connect(self._pump)
        self.timer.start()

        # simple volume metering (optional)
        self.volume_emitter = volume_emitter
        self.vol_window_bytes = max(1, self.bytes_per_ms * 100)  # ~100 ms
        self.vol_buffer = bytearray()
        self.vol_timer = QTimer(self)
        self.vol_timer.setTimerType(Qt.PreciseTimer)
        self.vol_timer.setInterval(33)
        self.vol_timer.timeout.connect(self._emit_volume_tick)
        self.vol_timer.start()
        self._sf = fmt.sampleFormat()

        self.on_stopped = None  # callback set by NativeBackend

    def feed(self, data: bytes) -> None:
        """
        Feed PCM bytes (already in device format).

        :param data: bytes
        """
        if not data or self.io is None:
            return
        self.buffer.extend(data)
        # NOTE: try pump quickly (non-blocking)
        self._pump()

    def mark_final(self) -> None:
        """Mark no-more-data and add small silence tail."""
        if not self.final:
            pad = self._align_down(self.bytes_per_ms * self.tail_ms)
            if pad > 0:
                self.buffer.extend(self._silence(pad))
        self.final = True
        self._pump()

    def stop(self) -> None:
        """Stop the session and clean up."""
        try:
            if self.timer:
                self.timer.stop()
        except Exception:
            pass
        try:
            if self.vol_timer:
                self.vol_timer.stop()
        except Exception:
            pass
        try:
            if self.sink:
                self.sink.stop()
        except Exception:
            pass

        # zero volume instantly
        try:
            self.vol_buffer.clear()
            if self.volume_emitter:
                self.volume_emitter(0)
        except Exception:
            pass

        self.io = None
        self.sink = None

        cb = self.on_stopped
        self.on_stopped = None
        if cb:
            try:
                cb()
            except Exception:
                pass

        self.deleteLater()

    def _pump(self) -> None:
        """Write as much as device can take right now."""
        if self.io is None:
            return

        free = int(self.sink.bytesFree()) if self.sink else 0
        if free <= 0 and not self.buffer:
            return

        to_write = min(len(self.buffer), free, self.max_write_bytes)
        to_write = self._align_down(to_write)

        # avoid tiny writes unless finishing
        if not self.final and 0 < to_write < self.min_write_bytes:
            return

        if to_write > 0:
            chunk = bytes(self.buffer[:to_write])
            written = self.io.write(chunk)
            if written and written > 0:
                del self.buffer[:written]
                # simple volume window
                self._vol_push(chunk[:self._align_down(written)])

        # stop when: final AND our buffer empty AND device queue empty
        if self.final and not self.buffer:
            pend = 0
            try:
                pend = int(self.sink.bufferSize()) - int(self.sink.bytesFree())
            except Exception:
                pass
            if pend <= 0:
                self.stop()

    def _align_down(self, n: int) -> int:
        """
        Align to frame boundary.

        :param n: number of bytes
        :return: aligned number of bytes
        """
        if self.frame_bytes <= 1:
            return n
        rem = n % self.frame_bytes
        return n - rem

    def _silence(self, n: int) -> bytes:
        """
        Generate n bytes of silence.
        NOTE: for Int16 silence is all zeros; for UInt8 it is 0x80.

        :param n: number of bytes
        :return: bytes of silence
        """
        if n <= 0:
            return b""
        sf = self.format.sampleFormat()
        if sf == QAudioFormat.SampleFormat.UInt8:
            return bytes([128]) * n
        return b"\x00" * n

    def _vol_push(self, chunk: bytes) -> None:
        """
        Push chunk to volume buffer and trim if needed.

        :param chunk: bytes to push to volume buffer
        """
        if not chunk:
            return
        self.vol_buffer.extend(chunk)
        if len(self.vol_buffer) > self.vol_window_bytes:
            del self.vol_buffer[:len(self.vol_buffer) - self.vol_window_bytes]

    def _emit_volume_tick(self) -> None:
        """Emit volume level (0-100) based on current vol_buffer content."""
        if self.volume_emitter is None:
            return
        try:
            if not self.vol_buffer:
                self.volume_emitter(0)
                return

            sf = self._sf
            if sf == QAudioFormat.SampleFormat.UInt8:
                arr = np.frombuffer(self.vol_buffer, dtype=np.uint8).astype(np.int16)
                arr = (arr - 128).astype(np.float32) / 128.0
            elif sf == QAudioFormat.SampleFormat.Int16:
                arr = np.frombuffer(self.vol_buffer, dtype=np.int16).astype(np.float32) / 32768.0
            elif sf == QAudioFormat.SampleFormat.Int32:
                arr = np.frombuffer(self.vol_buffer, dtype=np.int32).astype(np.float32) / 2147483648.0
            elif sf == QAudioFormat.SampleFormat.Float:
                arr = np.frombuffer(self.vol_buffer, dtype=np.float32).astype(np.float32)
            else:
                arr = np.frombuffer(self.vol_buffer, dtype=np.int16).astype(np.float32) / 32768.0

            if arr.size == 0:
                self.volume_emitter(0)
                return

            rms = float(np.sqrt(np.mean(arr.astype(np.float64) ** 2)))
            db = -60.0 if rms <= 1e-9 else 20.0 * float(np.log10(min(1.0, rms)))
            db = max(-60.0, min(0.0, db))
            volume = int(((db + 60.0) / 60.0) * 100.0)
            self.volume_emitter(volume)
        except Exception:
            pass


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.08.14 03:00:00                  #
# ================================================== #

from typing import Dict, Any, Tuple

from pygpt_net.item.ctx import CtxItem
from pygpt_net.item.preset import PresetItem


class BaseAgent:
    def __init__(self, *args, **kwargs):
        self.id = ""
        self.type = ""
        self.mode = ""
        self.name = ""
        self.custom_id = None
        self.custom_options = None
        self.custom_schema = None

    def get_mode(self) -> str:
        """
        Return Agent mode

        :return: Agent mode
        """
        return self.mode

    def get_agent(
            self,
            window,
            kwargs: Dict[str, Any]
    ) -> Any:
        """
        Return Agent provider instance

        :param window: window instance
        :param kwargs: keyword arguments
        :return: Agent provider instance
        """
        pass

    def set_id(self, id: str):
        """
        Set custom ID for the agent

        :param id: Custom ID
        """
        self.custom_id = id

    def set_schema(self, schema: list):
        """
        Set custom schema for the agent

        :param schema: Custom schema
        """
        self.custom_schema = schema


    def set_options(self, options: dict):
        """
        Set custom options for the agent

        :param options: Custom options
        """
        self.custom_options = options

    def get_options(self) -> dict:
        """
        Return Agent options

        :return: Agent options
        """
        if self.custom_options is not None:
            return self.custom_options
        return {}

    async def run(
            self,
            window,
            agent_kwargs: Dict[str, Any] = None,
            previous_response_id: str = None,
            messages: list = None,
            ctx: CtxItem = None,
            stream: bool = False,
            bridge = None,
    ) -> Tuple[CtxItem, str, str]:
        """
        Run agent (async)

        :param window: Window instance
        :param agent_kwargs: Additional agent parameters
        :param previous_response_id: ID of the previous response (if any)
        :param messages: Conversation messages
        :param ctx: Context item
        :param stream: Whether to stream output
        :param bridge: Connection context for agent operations
        :return: Current ctx, final output, last response ID
        """
        pass

    def get_option(self, preset: PresetItem, section: str, key: str) -> Any:
        """
        Get specific option from preset

        :param preset: Preset item
        :param section: Section name
        :param key: Option key
        :return: Option value
        """
        if preset is None:
            print("No preset provided, returning default option value")
            return None
        extra = preset.extra
        agent_id = self.custom_id if self.custom_id is not None else self.id
        if not isinstance(extra, dict) or agent_id not in extra:
            return self.get_default(section, key)
        options = extra[agent_id]
        if section not in options:
            return self.get_default(section, key)
        if key not in options[section]:
            return self.get_default(section, key)
        option = options[section][key]
        if option is None:
            return self.get_default(section, key)
        return option


    def get_default(self, section: str, key: str) -> Any:
        """
        Get default option value

        :param section: Section name
        :param key: Option key
        :return: Default option value
        """
        options = self.get_options()
        if section not in options:
            return
        if key not in options[section]['options']:
            return
        return options[section]['options'][key].get('default', None)

    def get_default_prompt(self) -> str:
        """
        Get default prompt for the agent

        :return: Default prompt string
        """
        options = self.get_options()
        if '__prompt__' in options:
            return options['__prompt__']
        return ""


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.08.24 23:00:00                  #
# ================================================== #

from PySide6.QtCore import Qt
from PySide6.QtWidgets import QVBoxLayout, QLabel, QHBoxLayout, QWidget, QSplitter, QSizePolicy, QRadioButton, QCheckBox, QButtonGroup

from pygpt_net.ui.widget.calendar.select import CalendarSelect
from pygpt_net.ui.widget.element.checkbox import ColorCheckbox
from pygpt_net.ui.widget.element.labels import HelpLabel
from pygpt_net.ui.widget.textarea.calendar_note import CalendarNote
from pygpt_net.utils import trans


class Calendar:
    __slots__ = ('window',)

    def __init__(self, window=None):
        """
        Calendar UI

        :param window: Window instance
        """
        self.window = window

    def init(self):
        """
        Initialize painter

        :return: QWidget
        """
        ui = self.window.ui
        ui.calendar['select'] = CalendarSelect(self.window)
        ui.calendar['select'].setMinimumSize(200, 200)
        ui.calendar['select'].setGridVisible(True)
        ui.calendar['note'] = CalendarNote(self.window)

    def setup(self) -> QWidget:
        """
        Setup calendar

        :return: QWidget
        """
        self.init()
        body = self.window.core.tabs.from_widget(self.setup_calendar())
        body.append(self.window.ui.calendar['note'])
        body.append(self.window.ui.calendar['select'])
        return body

    def _on_filter_id_clicked(self, id_: int) -> None:
        key = "all" if id_ == 0 else ("pinned" if id_ == 1 else "indexed")
        self.window.controller.ctx.common.toggle_display_filter(key)

    def _on_counters_all_toggled(self, checked: bool) -> None:
        self.window.controller.calendar.note.toggle_counters_all(checked)

    def setup_filters(self) -> QWidget:
        """
        Setup calendar filters

        :return: QWidget
        """
        ui = self.window.ui
        nodes = ui.nodes

        label_existing = nodes.get('filter.ctx.label')
        if label_existing is None:
            layout = QHBoxLayout()
            widget = QWidget()
            rows = QVBoxLayout()
            widget.setLayout(rows)

            label = QLabel(trans("filter.ctx.label"), widget)

            radio_all = QRadioButton(trans("filter.ctx.radio.all"), widget)
            radio_pinned = QRadioButton(trans("filter.ctx.radio.pinned"), widget)
            radio_indexed = QRadioButton(trans("filter.ctx.radio.indexed"), widget)

            counters_all = QCheckBox(trans("filter.ctx.counters.all"), widget)

            layout.addWidget(label)
            layout.addWidget(radio_all)
            layout.addWidget(radio_pinned)
            layout.addWidget(radio_indexed)
            layout.addWidget(counters_all)
            layout.addStretch()

            nodes['filter.ctx.labels'] = ColorCheckbox(self.window)

            rows.addLayout(layout)
            rows.addWidget(nodes['filter.ctx.labels'])

            group = QButtonGroup(widget)
            group.setExclusive(True)
            group.addButton(radio_all, 0)
            group.addButton(radio_pinned, 1)
            group.addButton(radio_indexed, 2)
            group.idClicked.connect(self._on_filter_id_clicked)

            counters_all.toggled.connect(self._on_counters_all_toggled)

            nodes['filter.ctx.label'] = label
            nodes['filter.ctx.radio.all'] = radio_all
            nodes['filter.ctx.radio.pinned'] = radio_pinned
            nodes['filter.ctx.radio.indexed'] = radio_indexed
            nodes['filter.ctx.counters.all'] = counters_all
        else:
            widget = nodes['filter.ctx.label'].parentWidget()
            nodes['filter.ctx.label'].setText(trans("filter.ctx.label"))
            nodes['filter.ctx.radio.all'].setText(trans("filter.ctx.radio.all"))
            nodes['filter.ctx.radio.pinned'].setText(trans("filter.ctx.radio.pinned"))
            nodes['filter.ctx.radio.indexed'].setText(trans("filter.ctx.radio.indexed"))
            nodes['filter.ctx.counters.all'].setText(trans("filter.ctx.counters.all"))

        desired = bool(self.window.core.config.get("ctx.counters.all"))
        if nodes['filter.ctx.counters.all'].isChecked() != desired:
            nodes['filter.ctx.counters.all'].setChecked(desired)

        widget.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        return widget

    def setup_calendar(self) -> QWidget:
        """
        Setup calendar

        :return: QSplitter
        """
        ui = self.window.ui
        nodes = ui.nodes
        calendar = ui.calendar

        select_layout = QVBoxLayout()
        select_layout.addWidget(calendar['select'])
        select_layout.setContentsMargins(5, 0, 5, 0)

        tip = nodes.get('tip.output.tab.calendar')
        if tip is None:
            tip = HelpLabel(trans('tip.output.tab.calendar'), self.window)
            nodes['tip.output.tab.calendar'] = tip
        else:
            tip.setText(trans('tip.output.tab.calendar'))

        note_label = calendar.get('note.label')
        if note_label is None:
            note_label = QLabel(trans('calendar.note.label'))
            calendar['note.label'] = note_label
        else:
            note_label.setText(trans('calendar.note.label'))

        layout = QVBoxLayout()
        layout.addWidget(note_label)
        layout.addWidget(calendar['note'])
        layout.addWidget(tip)
        layout.setContentsMargins(0, 0, 0, 0)

        widget = QWidget()
        widget.setLayout(layout)

        filters = self.setup_filters()
        select_layout.addWidget(filters)

        select_widget = QWidget()
        select_widget.setLayout(select_layout)

        ui.splitters['calendar'] = QSplitter(Qt.Horizontal)
        ui.splitters['calendar'].addWidget(select_widget)
        ui.splitters['calendar'].addWidget(widget)
        ui.splitters['calendar'].setStretchFactor(0, 6)
        ui.splitters['calendar'].setStretchFactor(1, 4)

        ui.splitters['calendar'].setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        filters.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)

        return self.window.core.tabs.from_widget(ui.splitters['calendar'])


from typing import (
    TYPE_CHECKING,
    Any,
    AsyncGenerator,
    Dict,
    Generator,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
)

from ollama import AsyncClient, Client

from llama_index.core.base.llms.generic_utils import (
    achat_to_completion_decorator,
    astream_chat_to_completion_decorator,
    chat_to_completion_decorator,
    stream_chat_to_completion_decorator,
)
from llama_index.core.base.llms.types import (
    ChatMessage,
    ChatResponse,
    ChatResponseAsyncGen,
    ChatResponseGen,
    CompletionResponse,
    CompletionResponseAsyncGen,
    CompletionResponseGen,
    ImageBlock,
    LLMMetadata,
    MessageRole,
    TextBlock,
)
from llama_index.core.bridge.pydantic import Field, PrivateAttr
from llama_index.core.constants import DEFAULT_CONTEXT_WINDOW, DEFAULT_NUM_OUTPUTS
from llama_index.core.instrumentation import get_dispatcher
from llama_index.core.llms.callbacks import llm_chat_callback, llm_completion_callback
from llama_index.core.llms.function_calling import FunctionCallingLLM
from llama_index.core.llms.llm import ToolSelection, Model
from llama_index.core.program.utils import process_streaming_objects, FlexibleModel
from llama_index.core.prompts import PromptTemplate
from llama_index.core.types import PydanticProgramMode

if TYPE_CHECKING:
    from llama_index.core.tools.types import BaseTool

DEFAULT_REQUEST_TIMEOUT = 30.0
dispatcher = get_dispatcher(__name__)


def get_additional_kwargs(
    response: Dict[str, Any], exclude: Tuple[str, ...]
) -> Dict[str, Any]:
    return {k: v for k, v in response.items() if k not in exclude}


def force_single_tool_call(response: ChatResponse) -> None:
    tool_calls = response.message.additional_kwargs.get("tool_calls", [])
    if len(tool_calls) > 1:
        response.message.additional_kwargs["tool_calls"] = [tool_calls[0]]


class Ollama(FunctionCallingLLM):
    """
    Ollama LLM.

    Visit https://ollama.com/ to download and install Ollama.

    Run `ollama serve` to start a server.

    Run `ollama pull <name>` to download a model to run.

    Examples:
        `pip install llama-index-llms-ollama`

        ```python
        from llama_index.llms.ollama import Ollama

        llm = Ollama(model="llama2", request_timeout=60.0)

        response = llm.complete("What is the capital of France?")
        print(response)
        ```

    """

    base_url: str = Field(
        default="http://localhost:11434",
        description="Base url the model is hosted under.",
    )
    model: str = Field(description="The Ollama model to use.")
    temperature: Optional[float] = Field(
        default=None,
        description="The temperature to use for sampling.",
    )
    context_window: int = Field(
        default=-1,
        description="The maximum number of context tokens for the model.",
    )
    request_timeout: float = Field(
        default=DEFAULT_REQUEST_TIMEOUT,
        description="The timeout for making http request to Ollama API server",
    )
    prompt_key: str = Field(
        default="prompt", description="The key to use for the prompt in API calls."
    )
    json_mode: bool = Field(
        default=False,
        description="Whether to use JSON mode for the Ollama API.",
    )
    additional_kwargs: Dict[str, Any] = Field(
        default_factory=dict,
        description="Additional model parameters for the Ollama API.",
    )
    is_function_calling_model: bool = Field(
        default=True,
        description="Whether the model is a function calling model.",
    )
    keep_alive: Optional[Union[float, str]] = Field(
        default="5m",
        description="controls how long the model will stay loaded into memory following the request(default: 5m)",
    )

    _client: Optional[Client] = PrivateAttr()
    _async_client: Optional[AsyncClient] = PrivateAttr()

    def __init__(
        self,
        model: str,
        base_url: str = "http://localhost:11434",
        temperature: Optional[float] = None,
        context_window: int = -1,
        request_timeout: Optional[float] = DEFAULT_REQUEST_TIMEOUT,
        prompt_key: str = "prompt",
        json_mode: bool = False,
        additional_kwargs: Optional[Dict[str, Any]] = None,
        client: Optional[Client] = None,
        async_client: Optional[AsyncClient] = None,
        is_function_calling_model: bool = True,
        keep_alive: Optional[Union[float, str]] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            model=model,
            base_url=base_url,
            temperature=temperature,
            context_window=context_window,
            request_timeout=request_timeout,
            prompt_key=prompt_key,
            json_mode=json_mode,
            additional_kwargs=additional_kwargs or {},
            is_function_calling_model=is_function_calling_model,
            keep_alive=keep_alive,
            **kwargs,
        )

        self._client = client
        self._async_client = async_client

    @classmethod
    def class_name(cls) -> str:
        return "Ollama_llm"

    @property
    def metadata(self) -> LLMMetadata:
        """LLM metadata."""
        return LLMMetadata(
            context_window=self.get_context_window(),
            num_output=DEFAULT_NUM_OUTPUTS,
            model_name=self.model,
            is_chat_model=True,  # Ollama supports chat API for all models
            # TODO: Detect if selected model is a function calling model?
            is_function_calling_model=self.is_function_calling_model,
        )

    @property
    def client(self) -> Client:
        if self._client is None:
            self._client = Client(host=self.base_url, timeout=self.request_timeout)
        return self._client

    @property
    def async_client(self) -> AsyncClient:
        if self._async_client is None:
            self._async_client = AsyncClient(
                host=self.base_url, timeout=self.request_timeout
            )
        return self._async_client

    @property
    def _model_kwargs(self) -> Dict[str, Any]:
        base_kwargs = {
            "temperature": self.temperature,
            "num_ctx": self.get_context_window(),
        }
        return {
            **base_kwargs,
            **self.additional_kwargs,
        }

    def get_context_window(self) -> int:
        if self.context_window == -1:
            # Try to get the context window from the model info if not set
            info = self.client.show(self.model).modelinfo
            for key, value in info.items():
                if "context_length" in key:
                    self.context_window = int(value)
                    break

        # If the context window is still -1, use the default context window
        return self.context_window if self.context_window != -1 else DEFAULT_CONTEXT_WINDOW

    def _convert_to_ollama_messages(self, messages: Sequence[ChatMessage]) -> Dict:
        ollama_messages = []
        for message in messages:
            cur_ollama_message = {
                "role": message.role.value,
                "content": "",
            }
            for block in message.blocks:
                if isinstance(block, TextBlock):
                    cur_ollama_message["content"] += block.text
                elif isinstance(block, ImageBlock):
                    if "images" not in cur_ollama_message:
                        cur_ollama_message["images"] = []
                    cur_ollama_message["images"].append(
                        block.resolve_image(as_base64=True).read().decode("utf-8")
                    )
                else:
                    raise ValueError(f"Unsupported block type: {type(block)}")

            if "tool_calls" in message.additional_kwargs:
                cur_ollama_message["tool_calls"] = message.additional_kwargs[
                    "tool_calls"
                ]

            ollama_messages.append(cur_ollama_message)

        return ollama_messages

    def _get_response_token_counts(self, raw_response: dict) -> dict:
        """Get the token usage reported by the response."""
        try:
            prompt_tokens = raw_response["prompt_eval_count"]
            completion_tokens = raw_response["eval_count"]
            total_tokens = prompt_tokens + completion_tokens
        except KeyError:
            return {}
        except TypeError:
            return {}
        return {
            "prompt_tokens": prompt_tokens,
            "completion_tokens": completion_tokens,
            "total_tokens": total_tokens,
        }

    def _prepare_chat_with_tools(
        self,
        tools: List["BaseTool"],
        user_msg: Optional[Union[str, ChatMessage]] = None,
        chat_history: Optional[List[ChatMessage]] = None,
        verbose: bool = False,
        allow_parallel_tool_calls: bool = False,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        tool_specs = [
            tool.metadata.to_openai_tool(skip_length_check=True) for tool in tools
        ]

        if isinstance(user_msg, str):
            user_msg = ChatMessage(role=MessageRole.USER, content=user_msg)

        messages = chat_history or []
        if user_msg:
            messages.append(user_msg)

        return {
            "messages": messages,
            "tools": tool_specs or None,
        }

    def _validate_chat_with_tools_response(
        self,
        response: ChatResponse,
        tools: List["BaseTool"],
        allow_parallel_tool_calls: bool = False,
        **kwargs: Any,
    ) -> ChatResponse:
        """Validate the response from chat_with_tools."""
        if not allow_parallel_tool_calls:
            force_single_tool_call(response)
        return response

    def get_tool_calls_from_response(
        self,
        response: "ChatResponse",
        error_on_no_tool_call: bool = True,
    ) -> List[ToolSelection]:
        """Predict and call the tool."""
        if response.message.additional_kwargs.get("tool_calls", []) is None:
            response.message.additional_kwargs["tool_calls"] = []
        tool_calls = response.message.additional_kwargs.get("tool_calls", [])
        if len(tool_calls) < 1:
            if error_on_no_tool_call:
                raise ValueError(
                    f"Expected at least one tool call, but got {len(tool_calls)} tool calls."
                )
            else:
                return []

        tool_selections = []
        for tool_call in tool_calls:
            argument_dict = tool_call["function"]["arguments"]

            tool_selections.append(
                ToolSelection(
                    # tool ids not provided by Ollama
                    tool_id=tool_call["function"]["name"],
                    tool_name=tool_call["function"]["name"],
                    tool_kwargs=argument_dict,
                )
            )

        return tool_selections

    @llm_chat_callback()
    def chat(self, messages: Sequence[ChatMessage], **kwargs: Any) -> ChatResponse:
        ollama_messages = self._convert_to_ollama_messages(messages)

        tools = kwargs.pop("tools", None)
        format = kwargs.pop("format", "json" if self.json_mode else None)

        response = self.client.chat(
            model=self.model,
            messages=ollama_messages,
            stream=False,
            format=format,
            tools=tools,
            options=self._model_kwargs,
            keep_alive=self.keep_alive,
        )

        response = dict(response)

        tool_calls = response["message"].get("tool_calls", [])
        token_counts = self._get_response_token_counts(response)
        if token_counts:
            response["usage"] = token_counts

        return ChatResponse(
            message=ChatMessage(
                content=response["message"]["content"],
                role=response["message"]["role"],
                additional_kwargs={"tool_calls": tool_calls},
            ),
            raw=response,
        )

    @llm_chat_callback()
    def stream_chat(
        self, messages: Sequence[ChatMessage], **kwargs: Any
    ) -> ChatResponseGen:
        ollama_messages = self._convert_to_ollama_messages(messages)

        tools = kwargs.pop("tools", None)
        format = kwargs.pop("format", "json" if self.json_mode else None)

        def gen() -> ChatResponseGen:
            response = self.client.chat(
                model=self.model,
                messages=ollama_messages,
                stream=True,
                format=format,
                tools=tools,
                options=self._model_kwargs,
                keep_alive=self.keep_alive,
            )

            response_txt = ""
            seen_tool_calls = set()
            all_tool_calls = []

            for r in response:
                if r["message"]["content"] is None:
                    continue

                r = dict(r)

                response_txt += r["message"]["content"]

                # FIX: 
                if r["message"].get("tool_calls", []) is None:
                    r["message"]["tool_calls"] = []

                new_tool_calls = [dict(t) for t in r["message"].get("tool_calls", [])]
                for tool_call in new_tool_calls:
                    if (
                        str(tool_call["function"]["name"]),
                        str(tool_call["function"]["arguments"]),
                    ) in seen_tool_calls:
                        continue
                    seen_tool_calls.add(
                        (
                            str(tool_call["function"]["name"]),
                            str(tool_call["function"]["arguments"]),
                        )
                    )
                    all_tool_calls.append(tool_call)
                token_counts = self._get_response_token_counts(r)
                if token_counts:
                    r["usage"] = token_counts

                yield ChatResponse(
                    message=ChatMessage(
                        content=response_txt,
                        role=r["message"]["role"],
                        additional_kwargs={"tool_calls": list(set(all_tool_calls))},
                    ),
                    delta=r["message"]["content"],
                    raw=r,
                )

        return gen()

    @llm_chat_callback()
    async def astream_chat(
        self, messages: Sequence[ChatMessage], **kwargs: Any
    ) -> ChatResponseAsyncGen:
        ollama_messages = self._convert_to_ollama_messages(messages)

        tools = kwargs.pop("tools", None)
        format = kwargs.pop("format", "json" if self.json_mode else None)

        async def gen() -> ChatResponseAsyncGen:
            response = await self.async_client.chat(
                model=self.model,
                messages=ollama_messages,
                stream=True,
                format=format,
                tools=tools,
                options=self._model_kwargs,
                keep_alive=self.keep_alive,
            )

            response_txt = ""
            seen_tool_calls = set()
            all_tool_calls = []

            async for r in response:
                if r["message"]["content"] is None:
                    continue

                r = dict(r)

                response_txt += r["message"]["content"]

                new_tool_calls = [dict(t) for t in r["message"].get("tool_calls", [])]
                for tool_call in new_tool_calls:
                    if (
                        str(tool_call["function"]["name"]),
                        str(tool_call["function"]["arguments"]),
                    ) in seen_tool_calls:
                        continue
                    seen_tool_calls.add(
                        (
                            str(tool_call["function"]["name"]),
                            str(tool_call["function"]["arguments"]),
                        )
                    )
                    all_tool_calls.append(tool_call)
                token_counts = self._get_response_token_counts(r)
                if token_counts:
                    r["usage"] = token_counts

                yield ChatResponse(
                    message=ChatMessage(
                        content=response_txt,
                        role=r["message"]["role"],
                        additional_kwargs={"tool_calls": all_tool_calls},
                    ),
                    delta=r["message"]["content"],
                    raw=r,
                )

        return gen()

    @llm_chat_callback()
    async def achat(
        self, messages: Sequence[ChatMessage], **kwargs: Any
    ) -> ChatResponse:
        ollama_messages = self._convert_to_ollama_messages(messages)

        tools = kwargs.pop("tools", None)
        format = kwargs.pop("format", "json" if self.json_mode else None)

        response = await self.async_client.chat(
            model=self.model,
            messages=ollama_messages,
            stream=False,
            format=format,
            tools=tools,
            options=self._model_kwargs,
            keep_alive=self.keep_alive,
        )

        response = dict(response)

        tool_calls = response["message"].get("tool_calls", [])
        token_counts = self._get_response_token_counts(response)
        if token_counts:
            response["usage"] = token_counts

        return ChatResponse(
            message=ChatMessage(
                content=response["message"]["content"],
                role=response["message"]["role"],
                additional_kwargs={"tool_calls": tool_calls},
            ),
            raw=response,
        )

    @llm_completion_callback()
    def complete(
        self, prompt: str, formatted: bool = False, **kwargs: Any
    ) -> CompletionResponse:
        return chat_to_completion_decorator(self.chat)(prompt, **kwargs)

    @llm_completion_callback()
    async def acomplete(
        self, prompt: str, formatted: bool = False, **kwargs: Any
    ) -> CompletionResponse:
        return await achat_to_completion_decorator(self.achat)(prompt, **kwargs)

    @llm_completion_callback()
    def stream_complete(
        self, prompt: str, formatted: bool = False, **kwargs: Any
    ) -> CompletionResponseGen:
        return stream_chat_to_completion_decorator(self.stream_chat)(prompt, **kwargs)

    @llm_completion_callback()
    async def astream_complete(
        self, prompt: str, formatted: bool = False, **kwargs: Any
    ) -> CompletionResponseAsyncGen:
        return await astream_chat_to_completion_decorator(self.astream_chat)(
            prompt, **kwargs
        )

    @dispatcher.span
    def structured_predict(
        self,
        output_cls: Type[Model],
        prompt: PromptTemplate,
        llm_kwargs: Optional[Dict[str, Any]] = None,
        **prompt_args: Any,
    ) -> Model:
        if self.pydantic_program_mode == PydanticProgramMode.DEFAULT:
            llm_kwargs = llm_kwargs or {}
            llm_kwargs["format"] = output_cls.model_json_schema()

            messages = prompt.format_messages(**prompt_args)
            response = self.chat(messages, **llm_kwargs)

            return output_cls.model_validate_json(response.message.content or "")
        else:
            return super().structured_predict(
                output_cls, prompt, llm_kwargs, **prompt_args
            )

    @dispatcher.span
    async def astructured_predict(
        self,
        output_cls: Type[Model],
        prompt: PromptTemplate,
        llm_kwargs: Optional[Dict[str, Any]] = None,
        **prompt_args: Any,
    ) -> Model:
        if self.pydantic_program_mode == PydanticProgramMode.DEFAULT:
            llm_kwargs = llm_kwargs or {}
            llm_kwargs["format"] = output_cls.model_json_schema()

            messages = prompt.format_messages(**prompt_args)
            response = await self.achat(messages, **llm_kwargs)

            return output_cls.model_validate_json(response.message.content or "")
        else:
            return await super().astructured_predict(
                output_cls, prompt, llm_kwargs, **prompt_args
            )

    @dispatcher.span
    def stream_structured_predict(
        self,
        output_cls: Type[Model],
        prompt: PromptTemplate,
        llm_kwargs: Optional[Dict[str, Any]] = None,
        **prompt_args: Any,
    ) -> Generator[Union[Model, FlexibleModel], None, None]:
        """
        Stream structured predictions as they are generated.

        Args:
            output_cls: The Pydantic class to parse responses into
            prompt: The prompt template to use
            llm_kwargs: Optional kwargs for the LLM
            **prompt_args: Args to format the prompt with

        Returns:
            Generator yielding partial objects as they are generated

        """
        if self.pydantic_program_mode == PydanticProgramMode.DEFAULT:

            def gen(
                output_cls: Type[Model],
                prompt: PromptTemplate,
                llm_kwargs: Dict[str, Any],
                prompt_args: Dict[str, Any],
            ) -> Generator[Union[Model, FlexibleModel], None, None]:
                llm_kwargs = llm_kwargs or {}
                llm_kwargs["format"] = output_cls.model_json_schema()

                messages = prompt.format_messages(**prompt_args)
                response_gen = self.stream_chat(messages, **llm_kwargs)

                cur_objects = None
                for response in response_gen:
                    try:
                        objects = process_streaming_objects(
                            response,
                            output_cls,
                            cur_objects=cur_objects,
                            allow_parallel_tool_calls=False,
                            flexible_mode=True,
                        )
                        cur_objects = (
                            objects if isinstance(objects, list) else [objects]
                        )
                        yield objects
                    except Exception:
                        continue

            return gen(output_cls, prompt, llm_kwargs, prompt_args)
        else:
            return super().stream_structured_predict(
                output_cls, prompt, llm_kwargs, **prompt_args
            )

    @dispatcher.span
    async def astream_structured_predict(
        self,
        output_cls: Type[Model],
        prompt: PromptTemplate,
        llm_kwargs: Optional[Dict[str, Any]] = None,
        **prompt_args: Any,
    ) -> AsyncGenerator[Union[Model, FlexibleModel], None]:
        """Async version of stream_structured_predict."""
        if self.pydantic_program_mode == PydanticProgramMode.DEFAULT:

            async def gen(
                output_cls: Type[Model],
                prompt: PromptTemplate,
                llm_kwargs: Dict[str, Any],
                prompt_args: Dict[str, Any],
            ) -> AsyncGenerator[Union[Model, FlexibleModel], None]:
                llm_kwargs = llm_kwargs or {}
                llm_kwargs["format"] = output_cls.model_json_schema()

                messages = prompt.format_messages(**prompt_args)
                response_gen = await self.astream_chat(messages, **llm_kwargs)

                cur_objects = None
                async for response in response_gen:
                    try:
                        objects = process_streaming_objects(
                            response,
                            output_cls,
                            cur_objects=cur_objects,
                            allow_parallel_tool_calls=False,
                            flexible_mode=True,
                        )
                        cur_objects = (
                            objects if isinstance(objects, list) else [objects]
                        )
                        yield objects
                    except Exception:
                        continue

            return gen(output_cls, prompt, llm_kwargs, prompt_args)
        else:
            # Fall back to non-streaming structured predict
            return await super().astream_structured_predict(
                output_cls, prompt, llm_kwargs, **prompt_args
            )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.09.05 01:00:00                  #
# ================================================== #

from typing import Optional
from pygpt_net.core.bridge.context import MultimodalContext


class Audio:
    def __init__(self, window=None):
        """
        Audio helpers for Anthropic (currently no official input/output audio in Python SDK).

        :param window: Window instance
        """
        self.window = window

    def build_input_block(self, multimodal_ctx: Optional[MultimodalContext]) -> Optional[dict]:
        """
        Future hook: build input_audio block if Anthropic exposes it publicly.
        Currently returns None to avoid 400 errors.
        """
        return None


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.08.24 23:00:00                  #
# ================================================== #

from PySide6.QtGui import QAction, QIcon

from pygpt_net.utils import trans


class Audio:
    def __init__(self, window=None):
        """
        Menu setup

        :param window: Window instance
        """
        self.window = window

    def setup(self):
        """Setup audio menu"""
        w = self.window
        m = w.ui.menu

        title = trans("menu.audio")
        txt_output = trans("menu.audio.output")
        txt_input = trans("menu.audio.input")
        txt_ctrl_plugin = trans("menu.audio.control.plugin")
        txt_ctrl_global = trans("menu.audio.control.global")
        txt_cache_clear = trans("menu.audio.cache.clear")
        txt_stop = trans("menu.audio.stop")

        icon_cache = QIcon(":/icons/delete.svg")
        icon_stop = QIcon(":/icons/mute.svg")

        plugins_toggle = w.controller.plugins.toggle
        voice_toggle = w.controller.access.voice.toggle_voice_control
        audio_clear = w.controller.audio.clear_cache
        audio_stop = w.controller.audio.stop_audio

        menu = m.get('menu.audio')
        is_new_menu = menu is None
        if is_new_menu:
            menu = w.menuBar().addMenu(title)
            m['menu.audio'] = menu
        else:
            menu.setTitle(title)

        act_output = m.get('audio.output')
        if act_output is None:
            act_output = QAction(txt_output, w, checkable=True)
            m['audio.output'] = act_output
            act_output.triggered.connect(lambda _=False: plugins_toggle('audio_output'))
        else:
            act_output.setText(txt_output)

        act_input = m.get('audio.input')
        if act_input is None:
            act_input = QAction(txt_input, w, checkable=True)
            m['audio.input'] = act_input
            act_input.triggered.connect(lambda _=False: plugins_toggle('audio_input'))
        else:
            act_input.setText(txt_input)

        act_ctrl_plugin = m.get('audio.control.plugin')
        if act_ctrl_plugin is None:
            act_ctrl_plugin = QAction(txt_ctrl_plugin, w, checkable=True)
            m['audio.control.plugin'] = act_ctrl_plugin
            act_ctrl_plugin.triggered.connect(lambda _=False: plugins_toggle('voice_control'))
        else:
            act_ctrl_plugin.setText(txt_ctrl_plugin)

        act_ctrl_global = m.get('audio.control.global')
        if act_ctrl_global is None:
            act_ctrl_global = QAction(txt_ctrl_global, w, checkable=True)
            m['audio.control.global'] = act_ctrl_global
            act_ctrl_global.triggered.connect(lambda _=False: voice_toggle())
        else:
            act_ctrl_global.setText(txt_ctrl_global)

        act_cache_clear = m.get('audio.cache.clear')
        if act_cache_clear is None:
            act_cache_clear = QAction(icon_cache, txt_cache_clear, w, checkable=False)
            m['audio.cache.clear'] = act_cache_clear
            act_cache_clear.triggered.connect(lambda _=False: audio_clear())
        else:
            act_cache_clear.setText(txt_cache_clear)
            act_cache_clear.setIcon(icon_cache)

        act_stop = m.get('audio.stop')
        if act_stop is None:
            act_stop = QAction(icon_stop, txt_stop, w, checkable=False)
            m['audio.stop'] = act_stop
            act_stop.triggered.connect(lambda _=False: audio_stop())
        else:
            act_stop.setText(txt_stop)
            act_stop.setIcon(icon_stop)

        if is_new_menu:
            menu.addAction(act_input)
            menu.addAction(act_output)
            menu.addSeparator()
            menu.addAction(act_ctrl_plugin)
            menu.addAction(act_ctrl_global)
            menu.addSeparator()
            menu.addAction(act_cache_clear)
            menu.addAction(act_stop)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2026.01.05 20:00:00                  #
# ================================================== #

from typing import Optional, List, Dict, Any
import base64
import re


def _stringify_content(content) -> Optional[str]:
    """
    Convert various xAI content shapes into a plain text string.
    Handles:
    - str
    - list of parts (dicts with {'type':'text','text':...} or str)
    - objects with .text or .content attributes
    - dict with 'text' or nested shapes
    """
    try:
        if content is None:
            return None
        if isinstance(content, str):
            return content
        if isinstance(content, list):
            buf: List[str] = []
            for p in content:
                if isinstance(p, str):
                    buf.append(p)
                elif isinstance(p, dict):
                    if isinstance(p.get("text"), str):
                        buf.append(p["text"])
                    elif isinstance(p.get("content"), str):
                        buf.append(p["content"])
                    elif isinstance(p.get("delta"), str):
                        buf.append(p["delta"])
                else:
                    t = getattr(p, "text", None)
                    if isinstance(t, str):
                        buf.append(t)
            return "".join(buf) if buf else None
        if isinstance(content, dict):
            if isinstance(content.get("text"), str):
                return content["text"]
            if isinstance(content.get("content"), str):
                return content["content"]
            if isinstance(content.get("delta"), str):
                return content["delta"]
        t = getattr(content, "text", None)
        if isinstance(t, str):
            return t
        c = getattr(content, "content", None)
        if isinstance(c, str):
            return c
        return str(content)
    except Exception:
        return None


def _extract_http_urls_from_text(text: Optional[str]) -> List[str]:
    """
    Extract http(s) URLs from plain text.
    """
    if not text or not isinstance(text, str):
        return []
    pattern = re.compile(r"(https?://[^\s)>\]\"']+)", re.IGNORECASE)
    urls = pattern.findall(text)
    out, seen = [], set()
    for u in urls:
        if u not in seen:
            out.append(u)
            seen.add(u)
    return out


def _append_urls(ctx, state, urls: List[str]):
    """
    Merge a list of URLs into state.citations and ctx.urls (unique, http/https only).
    """
    if not urls:
        return
    if not isinstance(state.citations, list):
        state.citations = []
    if ctx.urls is None:
        ctx.urls = []
    seen = set(state.citations) | set(ctx.urls)
    for u in urls:
        if not isinstance(u, str):
            continue
        if not (u.startswith("http://") or u.startswith("https://")):
            continue
        if u in seen:
            continue
        state.citations.append(u)
        ctx.urls.append(u)
        seen.add(u)


def _try_save_data_url_image(core, ctx, data_url: str) -> Optional[str]:
    """
    Save data:image/*;base64,... to file and return path.
    """
    try:
        if not data_url.startswith("data:image/"):
            return None
        header, b64 = data_url.split(",", 1)
        ext = "png"
        if ";base64" in header:
            mime = header.split(";")[0].split(":")[1].lower()
            if "jpeg" in mime or "jpg" in mime:
                ext = "jpg"
            elif "png" in mime:
                ext = "png"
        img_bytes = base64.b64decode(b64)
        save_path = core.image.gen_unique_path(ctx, ext=ext)
        with open(save_path, "wb") as f:
            f.write(img_bytes)
        return save_path
    except Exception:
        return None


def _process_message_content_for_outputs(core, ctx, state, content):
    """
    Inspect assistant message content (list of parts) for image_url outputs and URLs.
    - If image_url.url is data:... -> save to file and append to state.image_paths + ctx.images
    - If image_url.url is http(s) -> append to ctx.urls
    - Extract URLs from adjacent text parts conservatively
    - If file part present -> auto-download via Files API
    """
    if not isinstance(content, list):
        return
    any_image = False
    for p in content:
        if not isinstance(p, dict):
            continue
        ptype = p.get("type")
        if ptype == "image_url":
            img = p.get("image_url") or {}
            url = img.get("url")
            if isinstance(url, str):
                if url.startswith("data:image/"):
                    path = _try_save_data_url_image(core, ctx, url)
                    if path:
                        if not isinstance(ctx.images, list):
                            ctx.images = []
                        ctx.images.append(path)
                        state.image_paths.append(path)
                        any_image = True
                elif url.startswith("http://") or url.startswith("https://"):
                    _append_urls(ctx, state, [url])
        elif ptype == "text":
            t = p.get("text")
            if isinstance(t, str):
                urls = _extract_http_urls_from_text(t)
                if urls:
                    _append_urls(ctx, state, urls)
        elif ptype == "file":
            fid = p.get("id") or p.get("file_id")
            if isinstance(fid, str):
                if not hasattr(state, "xai_downloaded_file_ids"):
                    state.xai_downloaded_file_ids = set()
                if fid not in state.xai_downloaded_file_ids:
                    try:
                        path = core.api.xai.store.download_to_dir(fid)
                    except Exception:
                        path = None
                    if path:
                        if not isinstance(ctx.files, list):
                            ctx.files = []
                        if path not in ctx.files:
                            ctx.files.append(path)
                        ext = path.lower().rsplit(".", 1)[-1] if "." in path else ""
                        if ext in ["png", "jpg", "jpeg", "gif", "bmp", "tiff", "webp"]:
                            if not isinstance(ctx.images, list):
                                ctx.images = []
                            if path not in ctx.images:
                                ctx.images.append(path)
                        state.xai_downloaded_file_ids.add(fid)
    if any_image:
        try:
            state.has_xai_inline_image = True
        except Exception:
            pass


def _merge_tool_calls(state, new_calls: List[Dict[str, Any]]):
    """
    Merge a list of tool_calls (dict-like) into state.tool_calls with de-duplication and streaming concat of arguments.
    """
    if not new_calls:
        return
    if not isinstance(state.tool_calls, list):
        state.tool_calls = []

    def _norm(tc: Dict[str, Any]) -> Dict[str, Any]:
        fn = tc.get("function") or {}
        args = fn.get("arguments")
        if isinstance(args, (dict, list)):
            try:
                import json as _json
                args = _json.dumps(args, ensure_ascii=False)
            except Exception:
                args = str(args)
        return {
            "id": tc.get("id") or "",
            "type": "function",
            "function": {
                "name": fn.get("name") or "",
                "arguments": args or "",
            },
        }

    def _find_existing(key_id: str, key_name: str) -> Optional[Dict[str, Any]]:
        if not state.tool_calls:
            return None
        for ex in state.tool_calls:
            if key_id and ex.get("id") == key_id:
                return ex
            if key_name and ex.get("function", {}).get("name") == key_name:
                # name match as fallback for SDKs that stream without ids
                return ex
        return None

    for tc in new_calls:
        if not isinstance(tc, dict):
            continue
        nid = tc.get("id") or ""
        fn = tc.get("function") or {}
        nname = fn.get("name") or ""
        nargs = fn.get("arguments")
        if isinstance(nargs, (dict, list)):
            try:
                import json as _json
                nargs = _json.dumps(nargs, ensure_ascii=False)
            except Exception:
                nargs = str(nargs)

        existing = _find_existing(nid, nname)
        if existing is None:
            state.tool_calls.append(_norm(tc))
        else:
            if nname:
                existing["function"]["name"] = nname
            if nargs:
                existing["function"]["arguments"] = (existing["function"].get("arguments", "") or "") + str(nargs)


def _maybe_collect_tail_meta(state, obj: Dict[str, Any], ctx=None):
    """
    Collect tail metadata like citations and usage into state (and ctx for urls), if these fields exist.
    """
    try:
        if not isinstance(obj, dict):
            return
        if "citations" in obj:
            c = obj.get("citations") or []
            if isinstance(c, list):
                if ctx is not None:
                    _append_urls(ctx, state, [u for u in c if isinstance(u, str)])
                else:
                    try:
                        setattr(state, "xai_stream_citations", c)
                    except Exception:
                        pass
        if "usage" in obj and isinstance(obj["usage"], dict):
            try:
                state.usage_vendor = "xai"
                u = obj["usage"]
                def _as_int(v):
                    try:
                        return int(v)
                    except Exception:
                        try:
                            return int(float(v))
                        except Exception:
                            return 0
                p = _as_int(u.get("prompt_tokens") or u.get("input_tokens") or 0)
                c = _as_int(u.get("completion_tokens") or u.get("output_tokens") or 0)
                r = _as_int(u.get("reasoning_tokens") or 0)
                t = _as_int(u.get("total_tokens") or (p + c + r))
                state.usage_payload = {"in": p, "out": max(0, t - p) if t else c, "reasoning": r, "total": t}
            except Exception:
                pass
    except Exception:
        pass


def process_xai_sdk_chunk(ctx, core, state, item) -> Optional[str]:
    """
    xAI SDK native streaming chunk.

    :param ctx: Chat context
    :param core: Core controller
    :param state: Chat state
    :param item: Incoming streaming chunk; supports:
                 - tuple(response, chunk) [old/new SDK style]
                 - chunk object with .delta/.content/.tool_calls/.tool_outputs/.citations
                 - dict/SimpleNamespace with OpenAI-like choices[0].delta.content
                 - dict event style with root 'delta' or 'message'
    :return: Extracted text delta or None
    """
    response = None
    chunk = None
    try:
        if isinstance(item, (list, tuple)) and len(item) == 2:
            response, chunk = item
        else:
            chunk = item
    except Exception:
        return None

    # persist last response and attach response id to ctx once
    try:
        if response is not None:
            state.xai_last_response = response
            rid = getattr(response, "id", None)
            if rid and not getattr(ctx, "msg_id", None):
                ctx.msg_id = str(rid)
                if not isinstance(ctx.extra, dict):
                    ctx.extra = {}
                ctx.extra["xai_response_id"] = ctx.msg_id
                try:
                    core.ctx.update_item(ctx)
                except Exception:
                    pass
    except Exception:
        pass

    # Citations at chunk-level (last chunk in live search)
    try:
        cites = getattr(chunk, "citations", None)
        if cites and isinstance(cites, list):
            _append_urls(ctx, state, [u for u in cites if isinstance(u, str)])
    except Exception:
        pass

    # Tool calls emitted as dedicated SDK objects
    try:
        tc_list = getattr(chunk, "tool_calls", None) or []
        if tc_list:
            norm_list = []
            for tc in tc_list:
                if tc is None:
                    continue
                fn = getattr(tc, "function", None)
                name = getattr(fn, "name", "") if fn is not None else ""
                args = getattr(fn, "arguments", "") if fn is not None else ""
                if isinstance(args, (dict, list)):
                    try:
                        import json as _json
                        args = _json.dumps(args, ensure_ascii=False)
                    except Exception:
                        args = str(args)
                norm_list.append({
                    "id": getattr(tc, "id", "") or "",
                    "type": "function",
                    "function": {"name": name or "", "arguments": args or ""},
                })
            if norm_list:
                _merge_tool_calls(state, norm_list)
                state.force_func_call = True
    except Exception:
        pass

    # Tool outputs: scan for URLs to enrich ctx.urls (best effort)
    try:
        to_list = getattr(chunk, "tool_outputs", None) or []
        for to in to_list:
            content = getattr(to, "content", None)
            if isinstance(content, str):
                _append_urls(ctx, state, _extract_http_urls_from_text(content))
    except Exception:
        pass

    # 1) Direct .content (SDK chunk)
    try:
        if hasattr(chunk, "content"):
            t = _stringify_content(getattr(chunk, "content"))
            if t:
                _append_urls(ctx, state, _extract_http_urls_from_text(t))
                return str(t)
    except Exception:
        pass

    # 2) .delta object or dict
    try:
        delta = getattr(chunk, "delta", None)
        if delta is not None:
            try:
                tc = delta.get("tool_calls") if isinstance(delta, dict) else getattr(delta, "tool_calls", None)
                if tc:
                    if isinstance(tc, list):
                        # already OpenAI-like dicts
                        _merge_tool_calls(state, tc)
                        state.force_func_call = True
            except Exception:
                pass

            dc = delta.get("content") if isinstance(delta, dict) else getattr(delta, "content", None)
            if dc is not None:
                t = _stringify_content(dc)
                if t:
                    _append_urls(ctx, state, _extract_http_urls_from_text(t))
                    return str(t)
            if isinstance(delta, str) and delta:
                _append_urls(ctx, state, _extract_http_urls_from_text(delta))
                return delta
    except Exception:
        pass

    # 3) OpenAI-like dict with choices[0].delta or choices[0].message
    try:
        if isinstance(chunk, dict):
            # tools/citations/usage meta
            try:
                chs = chunk.get("choices") or []
                if chs:
                    candidate = chs[0] or {}
                    tc = (candidate.get("delta") or {}).get("tool_calls") or candidate.get("tool_calls") or []
                    if tc:
                        _merge_tool_calls(state, tc)
                        state.force_func_call = True
            except Exception:
                pass

            # text delta/message content
            chs = chunk.get("choices") or []
            if chs:
                first = chs[0] or {}
                d = first.get("delta") or {}
                m = first.get("message") or {}
                if "content" in d and d["content"] is not None:
                    t = _stringify_content(d["content"])
                    if t:
                        _append_urls(ctx, state, _extract_http_urls_from_text(t))
                        return str(t)
                if "content" in m and m["content"] is not None:
                    mc = m["content"]
                    # inspect for image_url outputs and URLs
                    if isinstance(mc, list):
                        _process_message_content_for_outputs(core, ctx, state, mc)
                        out_parts: List[str] = []
                        for p in mc:
                            if isinstance(p, dict) and p.get("type") == "text":
                                t = p.get("text")
                                if isinstance(t, str):
                                    out_parts.append(t)
                        if out_parts:
                            txt = "".join(out_parts)
                            _append_urls(ctx, state, _extract_http_urls_from_text(txt))
                            return txt
                    else:
                        t = _stringify_content(mc)
                        if t:
                            _append_urls(ctx, state, _extract_http_urls_from_text(t))
                            return str(t)

            # root-level delta/message
            if isinstance(chunk.get("delta"), dict) and "content" in chunk["delta"]:
                t = _stringify_content(chunk["delta"]["content"])
                if t:
                    _append_urls(ctx, state, _extract_http_urls_from_text(t))
                    return str(t)
            if isinstance(chunk.get("message"), dict) and "content" in chunk["message"]:
                mc = chunk["message"]["content"]
                _process_message_content_for_outputs(core, ctx, state, mc if isinstance(mc, list) else [])
                if isinstance(mc, str):
                    _append_urls(ctx, state, _extract_http_urls_from_text(mc))
                    return mc

            # tail metadata: citations and usage
            _maybe_collect_tail_meta(state, chunk, ctx=ctx)
    except Exception:
        pass

    # 4) SimpleNamespace with choices[0].delta/message
    try:
        chs = getattr(chunk, "choices", None)
        if chs:
            first = chs[0]
            delta = getattr(first, "delta", None)
            message = getattr(first, "message", None)
            if delta is not None:
                try:
                    tc = getattr(delta, "tool_calls", None)
                    if tc:
                        _merge_tool_calls(state, tc if isinstance(tc, list) else [])
                        state.force_func_call = True
                except Exception:
                    pass
                c = getattr(delta, "content", None)
                if c is not None:
                    t = _stringify_content(c)
                    if t:
                        _append_urls(ctx, state, _extract_http_urls_from_text(t))
                        return str(t)
            if message is not None:
                c = getattr(message, "content", None)
                if c is not None:
                    if isinstance(c, list):
                        _process_message_content_for_outputs(core, ctx, state, c)
                        # optional: also extract text parts
                        out_parts: List[str] = []
                        for p in c:
                            if isinstance(p, dict) and p.get("type") == "text":
                                t = p.get("text")
                                if isinstance(t, str):
                                    out_parts.append(t)
                        if out_parts:
                            txt = "".join(out_parts)
                            _append_urls(ctx, state, _extract_http_urls_from_text(txt))
                            return txt
                    else:
                        t = _stringify_content(c)
                        if t:
                            _append_urls(ctx, state, _extract_http_urls_from_text(t))
                            return str(t)
    except Exception:
        pass

    # 5) Plain string
    if isinstance(chunk, str):
        _append_urls(ctx, state, _extract_http_urls_from_text(chunk))
        return chunk

    return None


def xai_extract_tool_calls(response) -> list[dict]:
    """
    Extract tool calls from xAI final response (proto or modern SDK/message shapes).

    :param response: xAI final response object or dict
    :return: List of tool calls in normalized dict format
    """
    out: list[dict] = []

    def _append_from_msg(msg_obj):
        try:
            if not msg_obj:
                return
            tcs = getattr(msg_obj, "tool_calls", None)
            if not tcs and isinstance(msg_obj, dict):
                tcs = msg_obj.get("tool_calls")
            if not tcs:
                return
            for tc in tcs:
                try:
                    fn = getattr(tc, "function", None)
                    if isinstance(tc, dict):
                        fn = tc.get("function", fn)
                    name = getattr(fn, "name", None) if fn is not None else None
                    args = getattr(fn, "arguments", None) if fn is not None else None
                    if isinstance(fn, dict):
                        name = fn.get("name", name)
                        args = fn.get("arguments", args)
                    if isinstance(args, (dict, list)):
                        try:
                            import json as _json
                            args = _json.dumps(args, ensure_ascii=False)
                        except Exception:
                            args = str(args)
                    out.append({
                        "id": (getattr(tc, "id", None) if not isinstance(tc, dict) else tc.get("id")) or "",
                        "type": "function",
                        "function": {"name": name or "", "arguments": args or "{}"},
                    })
                except Exception:
                    continue
        except Exception:
            pass

    try:
        if isinstance(response, dict):
            ch = (response.get("choices") or [])
            if ch:
                _append_from_msg(ch[0].get("message") or {})
            if "message" in response:
                _append_from_msg(response.get("message"))
            if "output_message" in response:
                _append_from_msg(response.get("output_message"))
            if out:
                return out
    except Exception:
        pass

    try:
        _append_from_msg(getattr(response, "message", None))
        _append_from_msg(getattr(response, "output_message", None))
        if out:
            return out
    except Exception:
        pass

    try:
        proto = getattr(response, "proto", None)
        if not proto:
            return out
        choices = getattr(proto, "choices", None) or []
        if not choices:
            return out
        msg = getattr(choices[0], "message", None)
        _append_from_msg(msg)
    except Exception:
        pass
    return out


def xai_extract_citations(response) -> list[str]:
    """
    Extract citations (URLs) from xAI final response if present.

    :param response: xAI final response object
    :return: List of citation URLs
    """
    def _norm_urls(raw) -> List[str]:
        urls: List[str] = []
        seen = set()
        if isinstance(raw, list):
            for it in raw:
                u = None
                if isinstance(it, str):
                    u = it
                elif isinstance(it, dict):
                    u = (it.get("url") or it.get("uri") or
                         (it.get("source") or {}).get("url") or (it.get("source") or {}).get("uri"))
                if isinstance(u, str) and (u.startswith("http://") or u.startswith("https://")):
                    if u not in seen:
                        urls.append(u); seen.add(u)
        return urls

    urls: list[str] = []
    try:
        cites = getattr(response, "citations", None)
        if cites is None and isinstance(response, dict):
            cites = response.get("citations")
        urls.extend([u for u in _norm_urls(cites or []) if u not in urls])
    except Exception:
        pass

    try:
        msg = getattr(response, "message", None)
        if msg is None and isinstance(response, dict):
            msg = response.get("message")
        if msg:
            mc = getattr(msg, "citations", None)
            if mc is None and isinstance(msg, dict):
                mc = msg.get("citations")
            urls.extend([u for u in _norm_urls(mc or []) if u not in urls])
    except Exception:
        pass

    try:
        out_msg = getattr(response, "output_message", None)
        if out_msg:
            mc = getattr(out_msg, "citations", None)
            if mc is None and isinstance(out_msg, dict):
                mc = out_msg.get("citations")
            urls.extend([u for u in _norm_urls(mc or []) if u not in urls])
    except Exception:
        pass

    try:
        proto = getattr(response, "proto", None)
        if proto:
            proto_cites = getattr(proto, "citations", None) or []
            urls.extend([u for u in _norm_urls(proto_cites) if u not in urls])
            choices = getattr(proto, "choices", None) or []
            if choices:
                m = getattr(choices[0], "message", None)
                if m:
                    urls.extend([u for u in _norm_urls(getattr(m, "citations", None) or []) if u not in urls])
    except Exception:
        pass
    return urls


def xai_extract_usage(response) -> dict:
    """
    Extract usage from xAI final response via proto or modern usage fields. Return {'in','out','reasoning','total'}.

    :param response: xAI final response object
    :return: Usage dict
    """
    def _as_int(v):
        try:
            return int(v)
        except Exception:
            try:
                return int(float(v))
            except Exception:
                return 0

    def _from_usage_dict(usage: Dict[str, Any]) -> dict:
        p = usage.get("prompt_tokens", usage.get("input_tokens", 0)) or 0
        c = usage.get("completion_tokens", usage.get("output_tokens", 0)) or 0
        r = usage.get("reasoning_tokens", 0) or 0
        t = usage.get("total_tokens")
        p = _as_int(p); c = _as_int(c); r = _as_int(r)
        t = _as_int(t if t is not None else (p + c + r))
        out_total = max(0, t - p) if t else c
        return {"in": p, "out": out_total, "reasoning": r, "total": t}

    if isinstance(response, dict):
        u = response.get("usage")
        if isinstance(u, dict):
            return _from_usage_dict(u)

    try:
        u = getattr(response, "usage", None)
        if isinstance(u, dict):
            return _from_usage_dict(u)
    except Exception:
        pass

    try:
        proto = getattr(response, "proto", None)
        usage = getattr(proto, "usage", None) if proto else None
        if usage:
            p = _as_int(getattr(usage, "prompt_tokens", 0) or 0)
            c = _as_int(getattr(usage, "completion_tokens", 0) or 0)
            r = _as_int(getattr(usage, "reasoning_tokens", 0) or 0)
            t = _as_int(getattr(usage, "total_tokens", (p + c + r)) or (p + c + r))
            out_total = max(0, t - p) if t else c
            return {"in": p, "out": out_total, "reasoning": r, "total": t}
    except Exception:
        pass

    return {}


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.08.26 01:00:00                  #
# ================================================== #

from typing import Dict, Any, List

from pygpt_net.core.bridge import BridgeContext
from pygpt_net.core.types import (
    AGENT_TYPE_LLAMA,
    AGENT_MODE_WORKFLOW,
)
from llama_index.core.llms.llm import LLM
from llama_index.core.tools.types import BaseTool

from pygpt_net.utils import trans
from .workflow.supervisor import (
    get_workflow,
    SUPERVISOR_PROMPT,
    WORKER_PROMPT,
)
from ..base import BaseAgent

class SupervisorAgent(BaseAgent):
    def __init__(self, *args, **kwargs):
        super(SupervisorAgent, self).__init__(*args, **kwargs)
        self.id = "supervisor"
        self.type = AGENT_TYPE_LLAMA
        self.mode = AGENT_MODE_WORKFLOW
        self.name = "Supervisor + worker"

    def get_agent(self, window, kwargs: Dict[str, Any]):
        """
        Get agent instance

        :param window: Window instance
        :param kwargs: Agent parameters
        :return: PlannerWorkflow instance
        """
        context = kwargs.get("context", BridgeContext())
        preset = context.preset
        tools: List[BaseTool] = kwargs.get("tools", []) or []
        llm_supervisor: LLM = kwargs.get("llm", None)
        verbose: bool = kwargs.get("verbose", False)
        max_steps: int = kwargs.get("max_steps", 12)

        # get prompts from options or use defaults
        prompt_supervisor = self.get_option(preset, "supervisor", "prompt")
        prompt_worker = self.get_option(preset, "worker", "prompt")
        if not prompt_supervisor:
            prompt_supervisor = SUPERVISOR_PROMPT
        if not prompt_worker:
            prompt_worker = WORKER_PROMPT

        # get worker LLM from options
        model_worker = window.core.models.get(
            self.get_option(preset, "worker", "model")
        )
        llm_worker = window.core.idx.llm.get(model_worker, stream=False)
        worker_memory_session_id = ""
        if context.ctx and context.ctx.meta:
            worker_memory_session_id = "llama_worker_session_" + str(context.ctx.meta.id)

        # create workflow
        return get_workflow(
                tools,
                llm_supervisor=llm_supervisor,
                llm_worker=llm_worker,
                verbose=verbose,
                max_steps=max_steps,
                prompt_supervisor=prompt_supervisor,
                prompt_worker=prompt_worker,
                worker_memory_session_id=worker_memory_session_id,
        )

    def get_options(self) -> Dict[str, Any]:
        """
        Return Agent options

        :return: dict of options
        """
        return {
            "supervisor": {
                "label": trans("agent.option.section.supervisor"),
                "options": {
                    "prompt": {
                        "type": "textarea",
                        "label": trans("agent.option.prompt"),
                        "description": trans("agent.option.prompt.supervisor.desc"),
                        "default": SUPERVISOR_PROMPT,
                    },
                }
            },
            "worker": {
                "label": trans("agent.option.section.worker"),
                "options": {
                    "model": {
                        "label": trans("agent.option.model"),
                        "type": "combo",
                        "use": "models",
                        "default": "gpt-4o",
                    },
                    "prompt": {
                        "type": "textarea",
                        "label": trans("agent.option.prompt"),
                        "description": trans("agent.option.prompt.worker.desc"),
                        "default": WORKER_PROMPT,
                    },
                }
            },
        }
